import javax.swing.event.MouseInputListener;
import java.awt.event.InputEvent;
import javax.swing.*;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.Point;
import java.awt.geom.GeneralPath;
import java.awt.Toolkit;
import java.awt.Stroke;
import java.awt.Shape;
import java.awt.BasicStroke;
import java.awt.Component;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;
import java.awt.image.*;
import java.util.List;
import java.util.ListIterator;
import java.util.LinkedList;
import java.util.Vector;
import rob.util.*;
import rob.exceptions.*;
import rob.math.geom.*;


public class PolygonManager implements MouseInputListener, RefreshListener//, KeyListener
{
	public static final int SELECTING = 1;
	public static final int GENERATING = 2;

	private int state = SELECTING;
	private StateControl stateControl = null;

	private boolean isGenerating = false;
	private GeneralPath polyUnderContruction = null;

	private Color colorPolygons = Color.blue;
	private Color colorSelectedPolygons = Color.red;

	private List listPolygons = new LinkedList();
	private List listSelectedPolygons = new LinkedList();

	private Vector vectPolygonListeners = new Vector(5, 5);

	private ViewGraphics actualCanvas;


	public PolygonManager()	
	{
		stateControl = new StateControl(null);
	}


	public void setState(int newState)
	{
		// Se havia algum polgono selecionado, o remove.
		if (newState == GENERATING && listSelectedPolygons.size() != 0)
		{
			Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
			g2D.setXORMode(Color.white);
			g2D.setColor(colorSelectedPolygons);
			g2D.draw((ShapePolygon) listSelectedPolygons.get(0));
			g2D.setXORMode(Color.white);
			g2D.setColor(colorPolygons);
			g2D.draw((ShapePolygon) listSelectedPolygons.get(0));
			g2D.dispose();
			PolygonEvent polyEvt = new PolygonEvent((ShapePolygon) listSelectedPolygons.get(0), this);
			ListIterator listItPolyListeners = vectPolygonListeners.listIterator();
			while(listItPolyListeners.hasNext())
				((PolygonListener) listItPolyListeners.next()).polygonDeselected(polyEvt);
			listSelectedPolygons.clear();
		}
		state = newState;
	}


	public int getState()			{return state;}


	public Action getAction()
	{
		return stateControl;
	}


	public void addPolygonListener(PolygonListener newListener)
	{
		vectPolygonListeners.add(newListener);
	}


	public void removePolygonListener(PolygonListener newListener)
	{
		vectPolygonListeners.remove(newListener);
	}


	public void mouseDragged(MouseEvent e)	{}
	public void mouseMoved(MouseEvent e) 	{}
	public void mouseClicked(MouseEvent e) 	{}
	public void mousePressed(MouseEvent e) 	{}
	public void mouseEntered(MouseEvent e) 	{}
	public void mouseExited(MouseEvent e) 	{}


	public void mouseReleased(MouseEvent e)
	{
		// por enquanto s seleciona um polgono de cada vez.
		if (state == SELECTING)
		{
			isGenerating = false;

			// Se havia algum polgono selecionado, o remove.
			if (listSelectedPolygons.size() != 0)
			{
				Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
				g2D.setXORMode(Color.white);
				g2D.setColor(colorSelectedPolygons);
				g2D.draw((ShapePolygon) listSelectedPolygons.get(0));
				g2D.setColor(colorPolygons);
				g2D.draw((ShapePolygon) listSelectedPolygons.get(0));
				g2D.dispose();
				PolygonEvent polyEvt = new PolygonEvent((ShapePolygon) listSelectedPolygons.get(0), this);
				ListIterator listItPolyListeners = vectPolygonListeners.listIterator();
				while(listItPolyListeners.hasNext())
					((PolygonListener) listItPolyListeners.next()).polygonDeselected(polyEvt);
				listSelectedPolygons.clear();
			}
			ShapePolygon polyActualSelected = getClickedPolygon(e.getPoint());

			// Se algum polgono foi selecionado, o adiciona.
			if (polyActualSelected != null)
			{
				Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
				g2D.setXORMode(Color.white);
				g2D.setColor(colorSelectedPolygons);
				g2D.draw(polyActualSelected);
				g2D.dispose();
				listSelectedPolygons.add(polyActualSelected);
				PolygonEvent polyEvt = new PolygonEvent(polyActualSelected, this);
				ListIterator listItPolyListeners = vectPolygonListeners.listIterator();
				while(listItPolyListeners.hasNext())
					((PolygonListener) listItPolyListeners.next()).polygonSelected(polyEvt);
			}
		}
		else if (state == GENERATING)
		{
			isGenerating = true;
			if (polyUnderContruction == null)
			{
				polyUnderContruction = new GeneralPath();
				polyUnderContruction.moveTo((float) e.getPoint().getX(), (float) e.getPoint().getY());
			}
			else
			{
				Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
				g2D.setXORMode(Color.white);
				g2D.setColor(colorSelectedPolygons);
				g2D.draw(polyUnderContruction);
				// Se o boto direito do mouse  pressionado, acabar o polgono.
				if ((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0)
				{
					polyUnderContruction.closePath();
					ShapePolygon polyNew = new ShapePolygon(polyUnderContruction);
					if (polyNew.size() <= 2 || !polyNew.isPolygon())
						JOptionPane.showMessageDialog(null, "No  um polgono.");
					else
					{
						g2D.setColor(colorPolygons);
						g2D.draw(polyNew);
						listPolygons.add(polyNew);
						PolygonEvent polyEvt = new PolygonEvent(polyNew, this);
						ListIterator listItPolyListeners = vectPolygonListeners.listIterator();
						while(listItPolyListeners.hasNext())
							((PolygonListener) listItPolyListeners.next()).polygonAdded(polyEvt);
					}
					polyUnderContruction = null;
					isGenerating = false;
					state = SELECTING;
					stateControl.putValue(Action.NAME, "SELECTING");
				}

				// Se o boto direito do mouse no  pressionado, continuar o polgono.
				else
				{
					polyUnderContruction.lineTo((float) e.getPoint().getX(), (float) e.getPoint().getY());
					g2D.draw(polyUnderContruction);
				}
				g2D.dispose();
			}
		}
	}


	protected ShapePolygon getClickedPolygon(Point pointClicked)
	{
		Rectangle rectTmp = new Rectangle((int) pointClicked.getX()-2, (int) pointClicked.getY()-2, 4, 4);
		ShapePolygon polyTmp = null;
		ListIterator listItPolygons = listPolygons.listIterator();
		while (listItPolygons.hasNext())
		{
			polyTmp = (ShapePolygon) listItPolygons.next();
			if (polyTmp.intersects(rectTmp))
				return polyTmp;
		}
		return null;
	}


	public void addPolygon(List newPoly)
	{
		ShapePolygon polyNew = new ShapePolygon(newPoly);
		listPolygons.add(polyNew);
		Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
		g2D.setColor(colorPolygons);g2D.setXORMode(Color.white);
		g2D.draw(polyNew);
		PolygonEvent polyEvt = new PolygonEvent(polyNew, this);
		ListIterator listItPolyListeners = vectPolygonListeners.listIterator();
		while(listItPolyListeners.hasNext())
			((PolygonListener) listItPolyListeners.next()).polygonAdded(polyEvt);
		g2D.dispose();
	}


	public void removePolygon(IPolygon newPoly)
	{
		if (listPolygons.remove(newPoly))
		{
			Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
			g2D.setXORMode(Color.white);
			if (listSelectedPolygons.remove(newPoly))
				g2D.setColor(colorSelectedPolygons); 
			else
				g2D.setColor(colorPolygons); 
			g2D.draw((ShapePolygon) newPoly);
			PolygonEvent polyEvt = new PolygonEvent(newPoly, this);
			ListIterator listItPolyListeners = vectPolygonListeners.listIterator();
			while(listItPolyListeners.hasNext())
				((PolygonListener) listItPolyListeners.next()).polygonRemoved(polyEvt);
			g2D.dispose();
		}
	}


	public void addViewGraphics(ViewGraphics newCanvas)
	{
		actualCanvas = newCanvas;
	}


	public void removeViewGraphics(ViewGraphics newCanvas)
	{
		if (actualCanvas == newCanvas)
			actualCanvas = null;
	}



	public void refresh(RefreshEvent e)
	{
		Graphics2D g2D = (Graphics2D) actualCanvas.getGraphics();
		g2D.setColor(colorPolygons); g2D.setXORMode(Color.white);

		ListIterator listItPoly = listPolygons.listIterator();
		while(listItPoly.hasNext())
			g2D.draw((Shape) listItPoly.next());

		listItPoly = listSelectedPolygons.listIterator();
		while(listItPoly.hasNext())
			g2D.draw((Shape) listItPoly.next());

		g2D.setColor(colorSelectedPolygons);

		listItPoly = listSelectedPolygons.listIterator();
		while(listItPoly.hasNext())
			g2D.draw((Shape) listItPoly.next());
		g2D.dispose();
	}


	public class  StateControl extends AbstractControlAction
	{
		public StateControl(Component popi)
		{
			super("SELECTING", popi);
		}


		public void actionPerformed(ActionEvent evt)
		{
			switch (PolygonManager.this.getState())
			{
				case PolygonManager.SELECTING:
				{
					putValue(Action.NAME, "GENERATING");
					PolygonManager.this.setState(PolygonManager.GENERATING);
					break;
				}
				case PolygonManager.GENERATING:
				{
					putValue(Action.NAME, "SELECTING");
					PolygonManager.this.setState(PolygonManager.SELECTING);
					break;
				}
			}
		}
	}
}