package rob.math;

import rob.exceptions.*;

public abstract class FunctionsR
{
	public static final int DEFAULT_VALUE = -1;
	
	private static double maxItDefault = 40; 
	private static double maxIt = 40;

	private static double precisionDefault = 1e-6; 
	private static double precision = 1e-6;
	
	private FunctionsR()
	{
	}


/*****************
 * Funes de controle.
 ****************/

	public static boolean setMaxIterations(int i)
	{
		if (FunctionsR.DEFAULT_VALUE == i)
			maxIt = maxItDefault;
		else if (i < 1)
			return false;
		else
			maxIt = i;
		return true;		
	}
	public static boolean setPrecision(double i)
	{
		if (FunctionsR.DEFAULT_VALUE == (int)i)
			maxIt = maxItDefault;
		else if (i <= 0 || i < 2*MathConsts.infinitesimalDouble)
			return false;
		else
			maxIt = i;
		return true;		
	}


/**
 * Funo que tira o mdulo de um double.
 *@retorna o mdulo de um double.
 */
	public static double mod(double num)
	{
		if (num > 0) 
			return num;
		return -num;
	}

/**
 * Funo que um double num a um expoente inteiro exp.
 */
	public static double power(double num, int exp)
	{
		if (exp == 0)
			return 1;

		boolean invert;
		if (exp < 0)
		{	
			exp = (int) rob.math.FunctionsZ.mod(exp);
			invert = true;
		}
		else
			invert = false;

		long counter;
		double retValue, retValueTmp;
		retValue = 1;
		retValueTmp = num;
		counter = 1;
		while (counter <= exp)
		{
			if ((counter & exp) != 0)
				retValue *= retValueTmp;
			retValueTmp *= retValueTmp;
			counter = counter << 1;
		}

		if (invert)
			return inverse(retValue);
		return retValue;
	}

/**
 * A funcao fatorial calcula o fatorial de fator. Se fator no for um inteiro, ele  arredondado
 * para baixo.
 */
	public static double fatorial(double fator)
	{
		double retValue, actualNumber;
			actualNumber = 0;
		retValue = 1;
		while (actualNumber < fator)
		{
			actualNumber += 1;
			retValue *= actualNumber;
		}
		return retValue;
	}

/**
 * Funo que inverte um double.
 */
	public static double inverse(double num)
	{
		return 1/num;
	}
/**	
 * A funcao exp calcula e elevado a x.
 */
	public static double exp(double x)
	{		
		double exp, powerTmp;
		long counterPower, counterExp;
		exp = 0; counterExp = -1; powerTmp = precision*2;
		while (++counterExp < maxIt && powerTmp > precision)
		{
			powerTmp = 1; counterPower = 0;
			while (++counterPower <= counterExp)
				powerTmp *= x/counterPower;			
			exp += powerTmp;
		}
		return exp;
	}

/**	
 * A funcao log10 ainda nao foi implementada.
 */
	public static double log10(double num)
	{
		return 0;
	}

/**	
 * A funcao log2 ainda nao foi implementada.
 */
	public static double log2(double num)
	{
		return 0;
	}

/**	
 * A funcao ln cacula o logaritmo neperiano de x.
 */
	public static double ln(double x) throws WrongValueException
	{
		if (x < 0)
			throw new WrongValueException("Negative value in ln.");
		long counterPower, counterLn, limPower;
		double lnep, powerTmp, t;
		t = (x-1)/(x+1);
		lnep = 0; 
		counterLn = -1; 
		powerTmp = precision*2;
		while (++counterLn < maxIt && powerTmp > precision)
		{
			powerTmp = 1; counterPower = 0; limPower = 2*counterLn+1;
			while (++counterPower <= limPower)
				powerTmp *= t;			
			powerTmp = powerTmp/limPower;			
			lnep += powerTmp;
		}
		return lnep*2;
	}

/**	
 * A funcao cos calcula o cosseno de x.
 */
	public static double cos(double x)
	{		
		while (x > MathConsts.PI*2)
			x -= MathConsts.PI*2;
		while (x < -MathConsts.PI*2)
			x += MathConsts.PI*2;
		if (x > MathConsts.PI)
			x = MathConsts.PI*2 - x;
		if (x < -MathConsts.PI)
			x = -x - MathConsts.PI*2;
		double exp, powerTmp;
		long counterPower, counterExp, limPower;
		exp = 0; counterExp = -1; powerTmp = precision*2;
		while (++counterExp < maxIt && mod(powerTmp) > precision)
		{
			powerTmp = 1; counterPower = 0;
			while (++counterPower <= counterExp)
				powerTmp *= (-1)*x/counterPower;			
			limPower = counterExp<<1; --counterPower;
			while (++counterPower <= limPower)
				powerTmp *= x/counterPower;			
			exp += powerTmp;
		}
		return exp;
	}

//  A funcao sen calcula o seno de x
	public static double sen(double x)
	{
		while (x > MathConsts.PI*2)
			x -= MathConsts.PI*2;
		while (x < -MathConsts.PI*2)
			x += MathConsts.PI*2;
		if (x > MathConsts.PI/2)
			x = MathConsts.PI - x;
		if (x < -MathConsts.PI/2)
			x = -x - MathConsts.PI;
		double exp, powerTmp;
		long counterPower, counterExp, limPower;
		exp = 0; counterExp = -1; powerTmp = precision*2;
		while (++counterExp < maxIt && mod(powerTmp) > precision)
		{
			powerTmp = 1; counterPower = 0;
			while (++counterPower <= counterExp)
				powerTmp *= (-1)*x/counterPower;			
			limPower = (counterExp<<1) + 1;  --counterPower;
			while (++counterPower <= limPower)
				powerTmp *= x/counterPower;			
			exp += powerTmp;
		}
		return exp;
	}

//  A funcao tg calcula a tangente de x
	public static double tg(double x)
	{
		double exp1, exp2, powerTmp;
		long counterPower, counterExp, limPower;
		exp1 = 0; exp2 = 0; counterExp = -1; powerTmp = precision*2;
		while (++counterExp < maxIt && mod(powerTmp) > precision)
		{
			powerTmp = 1; counterPower = 0;
			while (++counterPower <= counterExp)
				powerTmp *= (-1)*x/counterPower;			
			limPower = (counterExp<<1); --counterPower;
			while (++counterPower <= limPower)
				powerTmp *= x/counterPower;			
			exp1 += powerTmp;				// cosseno.
			exp2 += powerTmp*x/counterPower;	// seno.
		}
		return exp2/exp1;
	}
 
//  A funcao arctg calcula o arco tangente de x
	public static double arcTg(double x) 
	{
		boolean xMaiorQueUm;
		long counterPower, counterArcTg, limPower;
		double arcTgRetvalue, powerTmp;
		arcTgRetvalue = 0; counterArcTg = -1; powerTmp = precision*2;
		if (x > 1 || x < -1)
		{
			x = inverse(x);
			xMaiorQueUm = true;
		}
		else
			xMaiorQueUm = false;
		while (++counterArcTg < maxIt && mod(powerTmp) > precision)
		{
			powerTmp = 1; counterPower = 0;
			while (++counterPower <= counterArcTg)
				powerTmp *= (-1)*x;			
			limPower = (counterArcTg<<1) + 1; --counterPower;
			while (++counterPower <= limPower)
				powerTmp *= x;			
			powerTmp = powerTmp/limPower;			
			arcTgRetvalue += powerTmp;
		}
		if (xMaiorQueUm)
			if (arcTgRetvalue >= 0)
				return MathConsts.PI/2 - arcTgRetvalue;
			else
				return -MathConsts.PI/2 - arcTgRetvalue;
		return arcTgRetvalue;
	}

	public static double arcCos(double x)
	{
		return Math.acos(x);
	}

	public static double sqrt(double num)
	{
		return Math.sqrt(num);
	}

	public static boolean sinalTrocado(double num1, double num2)
	{
		if ((num1 >= 0 && num2 >= 0) || (num1 <= 0 && num2 <= 0))
			return false;
		else
			return true;
	}


}