package rob.math.geom;


import java.util.Comparator;
import java.awt.geom.*;
import rob.util.*;


public class EventStruct extends AVLTerminatedTree
{
	public static final int BEGIN = 4;
	public static final int CROSS = 2;
	public static final int END = 1;


	protected EventStruct.EventStructKey keyLast = null;


	public EventStruct()
	{
		super(new EventStructComparatorProp(), new EventStructResolutor(), new AVLTerminatedTree.TerminatorFactory());
	}


	public EventStruct(int improp)
	{
		super(new EventStructComparatorImprop(), new EventStructResolutor(), new AVLTerminatedTree.TerminatorFactory());
	}


	public void insertEvent(AVLTerminatedTree.AVLNode thisNode)
	{
		addNode(thisNode);
	}


	public void insertSegment(Point2D p1, Point2D p2)
	{
		AVLTerminatedTree.AVLNode nodeInseted = this.createEventStructNode(p1, p2, 0);
		addNode(nodeInseted);
		addNode(this.createEventStructNode(nodeInseted));
	}


	public void insertSegment(Point2D p1, Point2D p2, int owner)
	{
		AVLTerminatedTree.AVLNode nodeInseted = this.createEventStructNode(p1, p2, owner);
		addNode(nodeInseted);
		addNode(this.createEventStructNode(nodeInseted));
	}


	public void insertSegment(Line2D line)
	{
		AVLTerminatedTree.AVLNode nodeInseted = this.createEventStructNode(line.getP1(), line.getP2(), 0);
		addNode(nodeInseted);
		addNode(this.createEventStructNode(nodeInseted));
	}


	public void insertSegment(Line2D line, int owner)
	{
		AVLTerminatedTree.AVLNode nodeInseted = this.createEventStructNode(line.getP1(), line.getP2(), owner);
		addNode(nodeInseted);
		addNode(this.createEventStructNode(nodeInseted));
	}


	public AVLTerminatedTree.AVLNode getNextAndDelete()
	{
		if (head.left == null || head.left instanceof Terminator)
			return null;
		AVLTerminatedTree.AVLNode thisNode = head.left;
		while (!(thisNode.left instanceof Terminator))
			thisNode = thisNode.left;
		deleteNode(thisNode);
		keyLast = (EventStruct.EventStructKey) thisNode.key;
		return thisNode;	
	}


	// Cria n de insero.
	public AVLTerminatedTree.AVLNode createEventStructNode(Point2D pA, Point2D pB, int owner)
	{
		return new AVLTerminatedTree.AVLNode(new EventStructKey(pA, pB, owner), null);
	}


	// Cria n de remoo.
	public AVLTerminatedTree.AVLNode createEventStructNode(AVLTerminatedTree.AVLNode toBeRemoved)
	{
		return new AVLTerminatedTree.AVLNode(new EventStructKey(((EventStruct.EventStructKey)toBeRemoved.key).sweepNode), null);
	}


	// Cria n de interseco.
	public AVLTerminatedTree.AVLNode createEventStructNode(Point2D pEvt, AVLTerminatedTree.AVLNode lowerOne, AVLTerminatedTree.AVLNode upperOne)
	{
		EventStructKey keyNew = new EventCrossKey(pEvt, lowerOne, upperOne);

		// keyNew deve necessariamente ser maior do que keyLast para evitar loops infinitos.
		if (keyLast == null || comp.compare(keyNew, keyLast) > 0)
			return new AVLTerminatedTree.AVLNode(keyNew, null);
		else
			return null;
	}


	public static class EventStructKey
	{
		public int type = 0;
		public Point2D pEvt = null;
		public AVLTerminatedTree.AVLNode sweepNode = null;


		protected EventStructKey()
		{
		}


		public EventStructKey(Point2D pA, Point2D pB, int ownerPolygon)
		{
			type = EventStruct.BEGIN;
			sweepNode = SweepLine.createSweepLineNode(pA, pB, ownerPolygon);
			this.pEvt = ((SweepLine.SweepLineKey) sweepNode.key).p1;
		}


		public EventStructKey(AVLTerminatedTree.AVLNode toBeRemoved)
		{
			type = EventStruct.END;
			sweepNode = toBeRemoved;
			this.pEvt = ((SweepLine.SweepLineKey) sweepNode.key).p2;
		}


		public String toString()
		{
			return "(" + pEvt.getX() + ", " + pEvt.getY() + ")";
		}
	}


	public static class EventCrossKey extends EventStructKey
	{
		public AVLTerminatedTree.AVLNode successorNode = null;
		public AVLTerminatedTree.AVLNode antecessorNode = null;


		public EventCrossKey(Point2D pEvt, AVLTerminatedTree.AVLNode lowerOne, AVLTerminatedTree.AVLNode upperOne)
		{
			super();
			type = EventStruct.CROSS;
			this.pEvt = pEvt;
			successorNode = upperOne;
			antecessorNode = lowerOne;
		}
	}


/**
 *	Caso se deseje localizar interseces prprias.
 */
	public static class EventStructComparatorProp implements Comparator
	{
		// O evento sendo inserido  sempre o1.
		public int compare(Object o1, Object o2)
		{
			EventStructKey key1 = (EventStructKey) o1;
			EventStructKey key2 = (EventStructKey) o2;
			
			int result = GeomOperations.compareByX(key1.pEvt, key2.pEvt);
			if (result != 0)
				return result;
			else
			{
				if (key1.type != key2.type)
					return key1.type - key2.type;
				else
				{
					double area2 = 0;
					if (key1.type == EventStruct.CROSS)
					{
						// Verifica se qual dos segmentos antecessores dos eventos
						// snedo comparados est embaixo. Se o atual estiver mais
						// embaixo, para.
						EventCrossKey crossKey1 = (EventCrossKey) o1;
						EventCrossKey crossKey2 = (EventCrossKey) o2;
						area2 = GeomOperations.area2(((SweepLine.SweepLineKey)crossKey1.antecessorNode.key).p1, 
												((SweepLine.SweepLineKey)crossKey1.antecessorNode.key).p2, 
												((SweepLine.SweepLineKey)crossKey2.antecessorNode.key).p2);
						if (area2 != 0)
							return area2 > 0 ? 1 : -1;
						else
						{
							// Verifica se qual dos segmentos sucessores dos eventos
							// snedo comparados est embaixo. Se o atual estiver mais
							// embaixo, para.
							area2 = GeomOperations.area2(((SweepLine.SweepLineKey)crossKey1.successorNode.key).p1, 
													((SweepLine.SweepLineKey)crossKey1.successorNode.key).p2, 
													((SweepLine.SweepLineKey)crossKey2.successorNode.key).p2);
							if (area2 != 0)
								return area2 > 0 ? 1 : -1;
							else
								return 0;
						}
					}
					else
					{
						area2 = GeomOperations.area2(((SweepLine.SweepLineKey)key1.sweepNode.key).p1, 
												((SweepLine.SweepLineKey)key1.sweepNode.key).p2, 
												((SweepLine.SweepLineKey)key2.sweepNode.key).p2);
						if (area2 != 0)
							return area2 > 0 ? 1 : -1;
						else
						{
							area2 = GeomOperations.area2(((SweepLine.SweepLineKey)key1.sweepNode.key).p1, 
													((SweepLine.SweepLineKey)key1.sweepNode.key).p2, 
													((SweepLine.SweepLineKey)key2.sweepNode.key).p1);
							if (area2 != 0)
								return area2 > 0 ? 1 : -1;
							else
								return 0;
						}
					}
				}
			}
		}
	}


/**
 *	Caso se deseje localizar interseces imprprias.
 */
	public static class EventStructComparatorImprop implements Comparator
	{
		// O evento sendo inserido  sempre o1.
		public int compare(Object o1, Object o2)
		{
			EventStructKey key1 = (EventStructKey) o1;
			EventStructKey key2 = (EventStructKey) o2;
			
			int result = GeomOperations.compareByX(key1.pEvt, key2.pEvt);
			if (result != 0)
				return result;
			else
			{
				if (key1.type != key2.type)
					return key2.type - key1.type;
				else
				{
					double area2 = 0;
					if (key1.type == EventStruct.CROSS)
					{
						// Verifica se qual dos segmentos antecessores dos eventos
						// snedo comparados est embaixo. Se o atual estiver mais
						// embaixo, para.
						EventCrossKey crossKey1 = (EventCrossKey) o1;
						EventCrossKey crossKey2 = (EventCrossKey) o2;
						area2 = GeomOperations.area2(((SweepLine.SweepLineKey)crossKey1.antecessorNode.key).p1, 
												((SweepLine.SweepLineKey)crossKey1.antecessorNode.key).p2, 
												((SweepLine.SweepLineKey)crossKey2.antecessorNode.key).p2);
						if (area2 != 0)
							return area2 > 0 ? 1 : -1;
						else
						{
							// Verifica se qual dos segmentos sucessores dos eventos
							// snedo comparados est embaixo. Se o atual estiver mais
							// embaixo, para.
							area2 = GeomOperations.area2(((SweepLine.SweepLineKey)crossKey1.successorNode.key).p1, 
													((SweepLine.SweepLineKey)crossKey1.successorNode.key).p2, 
													((SweepLine.SweepLineKey)crossKey2.successorNode.key).p2);
							if (area2 != 0)
								return area2 > 0 ? 1 : -1;
							else
								return 0;
						}
					}
					else
					{
						area2 = GeomOperations.area2(((SweepLine.SweepLineKey)key1.sweepNode.key).p1, 
												((SweepLine.SweepLineKey)key1.sweepNode.key).p2, 
												((SweepLine.SweepLineKey)key2.sweepNode.key).p2);
						if (area2 != 0)
							return area2 > 0 ? 1 : -1;
						else
						{
							area2 = GeomOperations.area2(((SweepLine.SweepLineKey)key1.sweepNode.key).p1, 
													((SweepLine.SweepLineKey)key1.sweepNode.key).p2, 
													((SweepLine.SweepLineKey)key2.sweepNode.key).p1);
							if (area2 != 0)
								return area2 > 0 ? 1 : -1;
							else
								return 0;
						}
					}
				}
			}
		}
	}


	public static class EventStructResolutor implements Resolutor
	{		
		public int resolve(INodeMap nodeOne, INodeMap nodeTwo)
		{
			return 0;
		}


		public int resolveKeys(Object keyOne, Object keyTwo)
		{
			return 0;
		}
	}
}