package rob.math.geom;

import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import rob.math.*;
import rob.math.geom.*;

public class GeomOperations
{	
	protected GeomOperations()	{}


	// Retorna maior que zero se p1 > p2, menor se p1 < p2 e igual se p1 = p2.
	public static int compareByX(Point2D p1, Point2D p2)
	{
		if (p1.getX() > p2.getX())
			return 1;
		if (p1.getX() < p2.getX())
			return -1;
		if (p1.getY() > p2.getY())
			return 1;
		if (p1.getY() < p2.getY())
			return -1;
		return 0;
	}


	// Retorna maior que zero se p1 > p2, menor se p1 < p2 e igual se p1 = p2.
	public static int compareByY(Point2D p1, Point2D p2)
	{
		if (p1.getY() > p2.getY())
			return 1;
		if (p1.getY() < p2.getY())
			return -1;
		if (p1.getX() > p2.getX())
			return 1;
		if (p1.getX() < p2.getX())
			return -1;
		return 0;
	}


	public static int compareByAngle(Point2D p1, Point2D p2)
	{
		Point2D ptComp = new Point2D.Double(0, 0);
		double valComp = GeomOperations.area2(ptComp, p2, p1);
		if (valComp > 0)
			return 1;
		if (valComp < 0)
			return -1;
		return 0;
	}


 	public static double area2(Point2D p1, Point2D p2, Point2D pTest)
	{
		return p1.getX()*(p2.getY() - pTest.getY()) - p1.getY()*(p2.getX() - pTest.getX()) + p2.getX()*pTest.getY() - p2.getY()*pTest.getX();
	}


	public static boolean isLeft(Point2D p1, Point2D p2, Point2D pTest)
	{
		return (p1.getX()*(p2.getY() - pTest.getY()) - p1.getY()*(p2.getX() - pTest.getX()) + p2.getX()*pTest.getY() - p2.getY()*pTest.getX()) > 0;
	}


	public static boolean isLeft(Point2D p1, Point2D p2, double x, double y)
	{
		return p1.getX()*(p2.getY() - y) - p1.getY()*(p2.getX() - x) + p2.getX()*y - p2.getY()*x > 0;
	}


	public static boolean isLeft(double x1, double y1, double x2, double y2, double x3, double y3)
	{
		return (x1*(y2 - y3) - y1*(x2 - x3) + x2*y3 - y2*x3) > 0;
	}


	public static boolean isLeftOrOn(Point2D p1, Point2D p2, Point2D pTest)
	{
		return (p1.getX()*(p2.getY() - pTest.getY()) - p1.getY()*(p2.getX() - pTest.getX()) + p2.getX()*pTest.getY() - p2.getY()*pTest.getX()) >= 0;
	}

	
	public static boolean isCollinear(Point2D p1, Point2D p2, Point2D pTest)
	{
		return (p1.getX()*(p2.getY() - pTest.getY()) - p1.getY()*(p2.getX() - pTest.getX()) + p2.getX()*pTest.getY() - p2.getY()*pTest.getX()) == 0;
	}
	
	
/**
 *	Determina se o ponto pTest pertence propia ou impropriamente ao segmento [p1,p2].
 */
	public static boolean isBetween(Point2D p1, Point2D p2, Point2D pTest)
	{
		if (!isCollinear(p1, p2, pTest))
			return false;
		if (p1.getX() != p2.getX())
			return ((p1.getX() <= pTest.getX()) && (p2.getX() >= pTest.getX())) || ((p1.getX() >= pTest.getX()) && (p2.getX() <= pTest.getX()));
		return ((p1.getY() <= pTest.getY()) && (p2.getY() >= pTest.getY())) || ((p1.getY() >= pTest.getY()) && (p2.getY() <= pTest.getY()));
	}


/**
 *	Determina se as retas A e B se intersectam propiamente.
 */
	public static boolean intersect(Point2D pA1, Point2D pA2, Point2D pB1, Point2D pB2)
	{
		double A1xA2yMinA1yA2x = pA1.getX()*pA2.getY() - pA1.getY()*pA2.getX();
		double A1xB1yMinA1yB1x = pA1.getX()*pB1.getY() - pA1.getY()*pB1.getX();
		double A1xB2yMinA1yB2x = pA1.getX()*pB2.getY() - pA1.getY()*pB2.getX();
		double A2xB2yMinA2yB2x = pA2.getX()*pB2.getY() - pA2.getY()*pB2.getX();
		double A2xB1yMinA2yB1x = pA2.getX()*pB1.getY() - pA2.getY()*pB1.getX();
		double B1xB2yMinB1yB2x = pB1.getX()*pB2.getY() - pB1.getY()*pB2.getX();
		double A1A2B1 = A1xA2yMinA1yA2x + A2xB1yMinA2yB1x - A1xB1yMinA1yB1x;
		double A1A2B2 = A1xA2yMinA1yA2x + A2xB2yMinA2yB2x - A1xB2yMinA1yB2x;
		double B1B2A1 = B1xB2yMinB1yB2x - A1xB2yMinA1yB2x + A1xB1yMinA1yB1x;
		double B1B2A2 = B1xB2yMinB1yB2x - A2xB2yMinA2yB2x + A2xB1yMinA2yB1x;
		return ((A1A2B1 != 0) && (A1A2B2 != 0) && (B1B2A1 != 0) && (B1B2A2 != 0)) && ((A1A2B1 > 0) != (A1A2B2 > 0)) && ((B1B2A1 > 0) != (B1B2A2 > 0));
	}


	public static boolean intersect(Line2D l1, Line2D l2)
	{
		double A1xA2yMinA1yA2x = l1.getX1()*l1.getY2() - l1.getY1()*l1.getX2();
		double A1xB1yMinA1yB1x = l1.getX1()*l2.getY1() - l1.getY1()*l2.getX1();
		double A1xB2yMinA1yB2x = l1.getX1()*l2.getY2() - l1.getY1()*l2.getX2();
		double A2xB2yMinA2yB2x = l1.getX2()*l2.getY2() - l1.getY2()*l2.getX2();
		double A2xB1yMinA2yB1x = l1.getX2()*l2.getY1() - l1.getY2()*l2.getX1();
		double B1xB2yMinB1yB2x = l2.getX1()*l2.getY2() - l2.getY1()*l2.getX2();
		double A1A2B1 = A1xA2yMinA1yA2x + A2xB1yMinA2yB1x - A1xB1yMinA1yB1x;
		double A1A2B2 = A1xA2yMinA1yA2x + A2xB2yMinA2yB2x - A1xB2yMinA1yB2x;
		double B1B2A1 = B1xB2yMinB1yB2x - A1xB2yMinA1yB2x + A1xB1yMinA1yB1x;
		double B1B2A2 = B1xB2yMinB1yB2x - A2xB2yMinA2yB2x + A2xB1yMinA2yB1x;
		return ((A1A2B1 != 0) && (A1A2B2 != 0) && (B1B2A1 != 0) && (B1B2A2 != 0)) && ((A1A2B1 > 0) != (A1A2B2 > 0)) && ((B1B2A1 > 0) != (B1B2A2 > 0));
	}


/**
 *	Determina se as retas A e B se intersectam impropiamente.
 */
	public static boolean intersectImprop(Point2D pA1, Point2D pA2, Point2D pB1, Point2D pB2)
	{
		double A1xA2yMinA1yA2x = pA1.getX()*pA2.getY() - pA1.getY()*pA2.getX();
		double A1xB1yMinA1yB1x = pA1.getX()*pB1.getY() - pA1.getY()*pB1.getX();
		double A1xB2yMinA1yB2x = pA1.getX()*pB2.getY() - pA1.getY()*pB2.getX();
		double A2xB2yMinA2yB2x = pA2.getX()*pB2.getY() - pA2.getY()*pB2.getX();
		double A2xB1yMinA2yB1x = pA2.getX()*pB1.getY() - pA2.getY()*pB1.getX();
		double B1xB2yMinB1yB2x = pB1.getX()*pB2.getY() - pB1.getY()*pB2.getX();
		double A1A2B1 = A1xA2yMinA1yA2x + A2xB1yMinA2yB1x - A1xB1yMinA1yB1x;
		double A1A2B2 = A1xA2yMinA1yA2x + A2xB2yMinA2yB2x - A1xB2yMinA1yB2x;
		double B1B2A1 = B1xB2yMinB1yB2x - A1xB2yMinA1yB2x + A1xB1yMinA1yB1x;
		double B1B2A2 = B1xB2yMinB1yB2x - A2xB2yMinA2yB2x + A2xB1yMinA2yB1x;
		return ((A1A2B1 > 0) != (A1A2B2 > 0)) && ((B1B2A1 > 0) != (B1B2A2 > 0));
	}


	public static boolean intersectImprop(Line2D l1, Line2D l2)
	{
		double A1xA2yMinA1yA2x = l1.getX1()*l1.getY2() - l1.getY1()*l1.getX2();
		double A1xB1yMinA1yB1x = l1.getX1()*l2.getY1() - l1.getY1()*l2.getX1();
		double A1xB2yMinA1yB2x = l1.getX1()*l2.getY2() - l1.getY1()*l2.getX2();
		double A2xB2yMinA2yB2x = l1.getX2()*l2.getY2() - l1.getY2()*l2.getX2();
		double A2xB1yMinA2yB1x = l1.getX2()*l2.getY1() - l1.getY2()*l2.getX1();
		double B1xB2yMinB1yB2x = l2.getX1()*l2.getY2() - l2.getY1()*l2.getX2();
		double A1A2B1 = A1xA2yMinA1yA2x + A2xB1yMinA2yB1x - A1xB1yMinA1yB1x;
		double A1A2B2 = A1xA2yMinA1yA2x + A2xB2yMinA2yB2x - A1xB2yMinA1yB2x;
		double B1B2A1 = B1xB2yMinB1yB2x - A1xB2yMinA1yB2x + A1xB1yMinA1yB1x;
		double B1B2A2 = B1xB2yMinB1yB2x - A2xB2yMinA2yB2x + A2xB1yMinA2yB1x;
		return ((A1A2B1 > 0) != (A1A2B2 > 0)) && ((B1B2A1 > 0) != (B1B2A2 > 0));
	}


	public static Point2D intersection(Point2D p11, Point2D p12, Point2D p21, Point2D p22)
	{
		double x, y;
		if (p12.getX() - p11.getX() == 0)
		{
			x = p11.getX();
			y = (p22.getY() - p21.getY())/(p22.getX() - p21.getX())*(x - p21.getX()) + p21.getY();
		}
		else if (p22.getX() - p21.getX() == 0)
		{
			x = p21.getX();
			y = (p12.getY() - p11.getY())/(p12.getX() - p11.getX())*(x - p11.getX()) + p11.getY();
		}
		else
		{	
			double m1 = (p12.getY() - p11.getY())/(p12.getX() - p11.getX());
			double m2 = (p22.getY() - p21.getY())/(p22.getX() - p21.getX());
			x = (m1*p11.getX() - m2*p21.getX() + p21.getY() - p11.getY())/(m1 - m2);
			y = m2*x - m2*p21.getX() + p21.getY();
		}
		return new Point2D.Double(x, y);
	}
}