package rob.math.geom;


import java.util.List;
import java.awt.geom.Point2D;


public interface IPolygon extends List
{
/**
 *	Custo = o(n).
 */
	public int setVertexAsFirst(int index);
	public int remove(int fromIndex, int toIndex);


/**
 *	Custo = o(1).
 */
	public Point2D getPoint(int index);
	public double getX(int index);
	public double getY(int index);


/**
 *	Operaes de set abaixo jogam exceo se a mudana no criar um novo polgono. Essa
 * verificao extra custa O(n). Para no se preucupar com esse custo, crair e usar as funes
 * de um PolygonIterator associado.
 */
	public void setX(int index, double newValue);
	public void setY(int index, double newValue);
	public void setPoint(int index, Point2D newPoint);


/**
 *	Retorna um iterator para iterar pelos elementos do poligono. Um IPolygonIterator deve iterar
 * infinitamente por um polgono. As funes hasNext e hasPrevious indicam quando foi dada uma volta
 * no polgono, mas no bloqueiam next nem previous.
 */
	public IPolygonIterator polygonIterator();


/**
 *	Retorna dois poligonos sem destruir original se forem poligonos vlidos. Tempo O(n).
 */
	public int getSubPolygons(int fromIndex, int toIndex, IPolygon onePolygon, IPolygon otherPolygon);


/**
 *	Transforma o original em {fromIndex, ..., toIndex} e retorna outro {toIndex, ..., fromIndex}
 *  se forem poligonos vlidos. Tempo O(n).
 */
	public int split(int fromIndex, int toIndex, IPolygon otherPolygon);


/**
 *	Transforma o original em {fromIndex, ..., toIndex} Tempo O(n).
 */
	public int split(int fromIndex, int toIndex);


/**
 *	Transforma o original em {0, ..., toIndex} Tempo O(n).
 */
	public int split(int toIndex);


// Operaes especficas de polgono.

/**
 *	Custo = o(n).
 */
	public double getArea();

	public boolean isDiagonal(int index1, int index2);


/**
 *	Diz se o vrtice index a partir do atual  convexo. Tempo O(index).
 */
	public boolean isConvex(int index);


/**
 *	Diz se o polgono  convexo. Tempo O(n).
 */
	public boolean isConvex();


	public boolean isPolygon();

/**
 *	Consome tempo O(n).
 */
	public boolean intersects(Point2D p1, Point2D p2);
}