package rob.math.geom;


import javax.swing.JOptionPane;
import rob.util.*;
import rob.exceptions.*;
import java.util.Collection;
import java.util.List;
import java.util.ListIterator;
import java.util.ConcurrentModificationException;
import java.awt.Shape;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.PathIterator;
import java.awt.geom.AffineTransform;
import java.awt.Rectangle;
import java.awt.Point;


public class ShapePolygon extends CDLLPolygon implements Shape
{
	protected Point2D gravityCenter = null;	// speeds up some operations.

	public ShapePolygon()				{super();}
	public ShapePolygon(Collection points)	{super(points);}
	public ShapePolygon(Rectangle2D rect)	{super(rect);}
	public ShapePolygon(Shape shp)		{super(shp);}


/**
 *	Esta operao no requer que o polgono esteje orientado no sentido anti-horrio. 
 * Leva tempo O(n).
 */
	public boolean contains(double x, double y)
	{
		System.out.println("Cheevas Regal");
		int countCrosses = 0;
		Entry saveHead = head;
		Point2D p1;
		Point2D p2;
		do
		{
			p1 = (Point2D) saveHead.element;
			p2 = (Point2D) saveHead.next.element;
			if ((p1.getY() > y) != (p2.getY() > y))
				if (p1.getX() >= x || p2.getX() >= x)
				{
					if (p1.getX() >= x && p2.getX() >= x)
						countCrosses++;
					else
					{
						if(GeomOperations.isLeft(p1, p2, x, y))
						{
							if (p2.getY() > p1.getY())
								countCrosses++;
						}
						else
						{
							if (p2.getY() < p1.getY())
								countCrosses++;
						}
					}
				}
			saveHead = saveHead.next;
		}
		while (saveHead != head);
		System.out.println("CONTAINS_IND : countCrosses = " + countCrosses);
		return countCrosses%2 != 0;
	}


/**
 *	Consome tempo O(n).
 */
	public boolean contains(double x, double y, double w, double h)	{return !intersects(x, y, w, h) && contains(x, y);}
	public boolean contains(Point2D p)		{return contains(p.getX(), p.getY());}
	public boolean contains(Rectangle2D r)	{return contains(r.getX(), r.getY(), r.getWidth(), r.getHeight());}


/**
 *	Consome tempo O(n) se bounds ainda no esto calculados. Seno, consome O(1).
 */
	public Rectangle getBounds()
	{
		if (boundsModified)
			calculateBounds();
		return new Rectangle((int) xMin, (int) yMin,  (int) (xMax - xMin), (int) (yMax - yMin));
	}


	public Rectangle2D getBounds2D()	{return getBounds();}


	public PathIterator getPathIterator(AffineTransform at)			{return ShapePolygon.this.new CDLLPathIterator(at);}
	public PathIterator getPathIterator(AffineTransform at, double momy)	{return ShapePolygon.this.new CDLLPathIterator(at);}


/**
 *	Consome tempo O(n).
 */
	public boolean intersects(double x, double y, double w, double h)
	{
		Point2D pUpperLeft  = new Point((int) x, (int) y);
		Point2D pUpperRight = new Point((int) (x + w), (int) y);
		Point2D pLowerRight = new Point((int) (x + w), (int) (y + h));
		Point2D pLowerLeft  = new Point((int) x, (int) (y + h));
//////		return !intersects(pUpperLeft, pUpperRight) && !intersects(pUpperRight, pLowerRight) && !intersects(pLowerRight, pLowerLeft) && !intersects(pLowerLeft, pUpperLeft);
		return intersects(pUpperLeft, pUpperRight) || intersects(pUpperRight, pLowerRight) || intersects(pLowerRight, pLowerLeft) || intersects(pLowerLeft, pUpperLeft);
	}


/**
 *	Consome tempo O(n).
 */
	public boolean intersects(Rectangle2D r)	{return intersects(r.getX(), r.getY(), r.getWidth(), r.getHeight());}


	public class CDLLPathIterator implements PathIterator
	{
		private CDLLPolygonIterator doTheWork;
		boolean isDone = false;
		Point2D currentElement = null;
		boolean firstTime = true;
		private AffineTransform at;


		public CDLLPathIterator(AffineTransform at)
		{
/////////////			doTheWork = CDLLPolygon.this.new CDLLPolygonIterator();
			doTheWork = ShapePolygon.this.new CDLLPolygonIterator();
			if (at == null)
				this.at = new AffineTransform();
			else
				this.at = at;
		}


		public int getWindingRule()	{return PathIterator.WIND_EVEN_ODD;}
		public boolean isDone()		{return isDone;}


		public void next()
		{
			if (doTheWork.hasNext())
				currentElement = at.transform((Point2D) doTheWork.next(), currentElement);
			else
				currentElement = null;
		}


		public int currentSegment(float[] coords)
		{
			if (firstTime)
			{
				firstTime = false;
				if (doTheWork.hasNext())
				{
					currentElement = at.transform((Point2D) doTheWork.next(), currentElement);
					coords[0] = (float) currentElement.getX();
					coords[1] = (float) currentElement.getY();
					return PathIterator.SEG_MOVETO;
				}
				else
				{
					isDone = true;
					coords[0] = 0; coords[1] = 0;
					return PathIterator.SEG_CLOSE;
				}
			}
			else if (currentElement == null)
			{
				isDone = true;
				coords[0] = 500; coords[1] = 500;
				return PathIterator.SEG_CLOSE;
			}
			else
			{
				coords[0] = (float) currentElement.getX(); 
				coords[1] = (float) currentElement.getY();
				return PathIterator.SEG_LINETO;
			}
		}


		public int currentSegment(double[] coords)
		{
			if (firstTime)
			{
				firstTime = false;
				if (doTheWork.hasNext())
				{
					currentElement = (Point2D) doTheWork.next();
					coords[0] = currentElement.getX();
					coords[1] = currentElement.getY();
					return PathIterator.SEG_MOVETO;
				}
				else
				{
					isDone = true;
					coords[0] = 0; coords[1] = 0;
					return PathIterator.SEG_CLOSE;
				}
			}
			else if (currentElement == null)
			{
				isDone = true;
				coords[0] = 500; coords[1] = 500;
				return PathIterator.SEG_CLOSE;
			}
			else
			{
				coords[0] = currentElement.getX(); 
				coords[1] = currentElement.getY();
				return PathIterator.SEG_LINETO;
			}
		}
	}

///	Mtodos fora da interface Shape.

	public Point2D getGravityCenter()
	{
		if (gravityCenter != null)
			return gravityCenter;

		IPolygonIterator  polyIterator = this.polygonIterator();
		double x = 0, y = 0;
		while (polyIterator.hasNext())
		{
			polyIterator.next();
			x += polyIterator.getX();
			y += polyIterator.getY();
		}
		return new Point2D.Double(x/this.size(), y/this.size());
	}


	public ShapePolygon createTransformedShape(AffineTransform affTrf)
	{
		ShapePolygon shpRetVal = null;
		shpRetVal = new ShapePolygon((Shape) this);
		IPolygonIterator polyIt = shpRetVal.polygonIterator();
		while (polyIt.next() != null)
			affTrf.transform(polyIt.getPoint(), polyIt.getPoint());
		return shpRetVal;
	}
}