/*
 * @(#)TwoPolygonsOutputGenerator.java	1.0 2000/07/22
 *
 */
package rob.math.geom;

import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.util.Collection;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Comparator;
import rob.math.*;
import rob.util.*;
import rob.exceptions.*;


/**
 * Um objeto <code>TwoPolygonsOutputGenerator</code> produz e armazena em ordem
 * os vrtices que comporo o objeto de retorno, que varia dependendo do que est
 * sendfo retornado. No caso de achar apenas as interseces, se retorna uma lista
 * contendo os pontos de interseco. No caso de unio o retornado  uma lista 
 * contendo os pontos em ordem anti-horria j protnos para formarem um polgono.
 * E em caso de interseco  retornado umas lista contendo listas com pointos em
 * sentido anti-horrio contendo os polgonos decorrentes da interseco.
 * TwoPolygonsOutputGenerator  derivado de um tipo especial de rvore AVL, a 
 * <code>AVLTerminatedTree</code>.
 * A estrutura garante remoes, inseres e consultas em tempo O(lg(n)).
 * <p>
 *
 * @version 	1.0 2000/07/22
 * @author 	Roberto Giordano Barra
 * @see         
 */
public class TwoPolygonsOutputGenerator extends AVLTerminatedTree
{
	// Os tipos possveis de informao de ssida.
	public static final int CROSS_POINTS = 0;
	public static final int UNION = 1;
	public static final int INTERSECTION = 2;

	// O tipo selecionado para o objeto atual.
	private int selectedType = TwoPolygonsOutputGenerator.CROSS_POINTS;

	// lastEvent armazena o ltimo evento relevante que entrou na estrutura. Nele, so usadas as seguintes informaes:
	// lastEventKey.type, lastEventKey.pEvt, (TwoPolygonsSweepLine.SweepLineElement) lastEventKey.sweepNode.element
	private TwoPolygonsEventStruct.EventStructKey lastEventKey = null;
	private TwoPolygonsSweepLine.SweepLineElement lastSweepNodeElement = null;


	/**
	 * Cria um novo objeto do tipo type.
	 */
	public TwoPolygonsOutputGenerator(int type)
	{
		super(new TwoPolygonsOutputGenerator.OutputComparator(), new TwoPolygonsOutputGenerator.OutputResolutor());
		selectedType = type;		
	}


	/**
	 * Atualiza a estrutura com relao ao ponto-evento passado. A cadas ponto-evento adicionado  
 	 * decidido se deve ser adicionada uma novaentrada na lista ou se basta criar ligaes entre um 
 	 * vritce j existente ou se no se deve fazer nada.
	 * @param   actualEvent	o ponto-evento correspondente. 
	 */
	public void stepOutput(AVLTerminatedTree.AVLNode actualEvent)
	{
		// 1- Todos aceitam type = CROSS.
		// 2- S aceita type = END se (TwoPolygonsSweepLine.SweepLineElement) lastEventKey.sweepNode.element.vertex for != null (ou seja, j foi associado em um BEGIN anterior.)
		// 3- Se for type = BEGIN, cada qual tem suas regras.
		// 4 - RESTRIES: a) No ter mais de dois beigns ou ends coincidentes.
		//			 b) No ter crosses coincidentes.

		TwoPolygonsEventStruct.EventStructKey actualEventKey = (TwoPolygonsEventStruct.EventStructKey) actualEvent.key;

		if (selectedType == TwoPolygonsOutputGenerator.CROSS_POINTS)
		{
			if (actualEventKey.type == TwoPolygonsEventStruct.CROSS)
				put(new Vertex(actualEventKey.pEvt), null);
		}
		else if (selectedType == TwoPolygonsOutputGenerator.UNION)
		{
			if (actualEventKey.type == TwoPolygonsEventStruct.BEGIN)
			{
				TwoPolygonsSweepLine.SweepLineElement actualSweepNodeElement = null;
				actualSweepNodeElement = (TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element;

				// Sendo esse ponto-evento no-coincidente com o passado ...
				if (lastEventKey == null || !lastEventKey.pEvt.equals(actualEventKey.pEvt))
				{
					int actualLeftOwner = ((TwoPolygonsRegionsMarkersTerminatorFactory.Element) actualEventKey.sweepNode.getLeftTerminator().element).getOwner();

					// Se a aresta estiver sobre uma regio VACUUM ...
					if (actualLeftOwner == TwoPolygonsCrossings.VACUUM)
					{
						Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
						vNewEntry.vBegin = vNewEntry;
						put(vNewEntry, null);
						actualSweepNodeElement.setVertex(vNewEntry);
						actualSweepNodeElement.setUpperChain(false);
						lastSweepNodeElement = actualSweepNodeElement; lastEventKey = actualEventKey;
					}

					// Se a aresta estiver sobre uma regio owner ...
					else if (actualLeftOwner == ((TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element).getOwner())
					{
						Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
						vNewEntry.vBegin = vNewEntry;
						put(vNewEntry, null);
						actualSweepNodeElement.setVertex(vNewEntry);
						actualSweepNodeElement.setUpperChain(true);
						lastSweepNodeElement = actualSweepNodeElement; lastEventKey = actualEventKey;
					}
				}

				// Tendo BEGIN igual o mesmo ponto-evento que o anterior, pode ser duas coisas:
				else if (lastEventKey != null && lastEventKey.pEvt.equals(actualEventKey.pEvt))
				{
					// Pode ser a parte de cima de um outro BEGIN ...
					if (actualEventKey.type == lastEventKey.type)
					{
						int actualLeftOwner = ((TwoPolygonsRegionsMarkersTerminatorFactory.Element) actualEventKey.sweepNode.getLeftTerminator().element).getOwner();
						int owner = ((TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element).getOwner();
						int notOwner = (owner == TwoPolygonsCrossings.A) ? TwoPolygonsCrossings.B : TwoPolygonsCrossings.A;
						if (actualLeftOwner == TwoPolygonsCrossings.VACUUM || actualLeftOwner == owner)
						{
							boolean lastLeftUpperChain = ((TwoPolygonsSweepLine.SweepLineElement) lastEventKey.sweepNode.element).getUpperChain();
							actualSweepNodeElement.setVertex(lastSweepNodeElement.getVertex());
							actualSweepNodeElement.setUpperChain(!lastLeftUpperChain);
						}
					}

					// ... ou pode ser a continuao de um END. Quem cria o vertex  o evento de END,
					// no o de BEGIN. Ele s associa o vrtice.
					else
					{
						actualSweepNodeElement.setVertex(lastSweepNodeElement.getVertex());
						actualSweepNodeElement.setUpperChain(lastSweepNodeElement.getUpperChain());
					}
				}
			}
			else if (actualEventKey.type == TwoPolygonsEventStruct.END)
			{
				TwoPolygonsSweepLine.SweepLineElement actualSweepNodeElement = (TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element;

				// Se o ponto-evento anterior for igual ao atual, ento quer dizer que foi um evento de END
				// fechando as cadeias, e essa ltima  a superior.
				if (lastEventKey != null && lastEventKey.pEvt.equals(actualEventKey.pEvt))
				{
					if (actualSweepNodeElement.getUpperChain())
					{
						lastSweepNodeElement.getVertex().next = actualSweepNodeElement.getVertex();
						actualSweepNodeElement.getVertex().previous = lastSweepNodeElement.getVertex();
					}
					else
					{
						lastSweepNodeElement.getVertex().previous = actualSweepNodeElement.getVertex();
						actualSweepNodeElement.getVertex().next = lastSweepNodeElement.getVertex();
					}
				}

				// Se o ponto-evento no for o mesmo mas o sweepNode.Element estiver associado, ento 
				// um novo vrtice a ser adicionado.
				else if (actualSweepNodeElement.getVertex() != null)
				{
					Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
					vNewEntry.vBegin = actualSweepNodeElement.getVertex().vBegin;

					// Se  pra adicionar na cadeia superior ...
					if (actualSweepNodeElement.getUpperChain())
					{
						actualSweepNodeElement.getVertex().previous = vNewEntry;
						vNewEntry.next = actualSweepNodeElement.getVertex();
					}

					// ... ou na cadeia inferior.
					else
					{
						actualSweepNodeElement.getVertex().next = vNewEntry;
						vNewEntry.previous = actualSweepNodeElement.getVertex();
					}
					actualSweepNodeElement.setVertex(vNewEntry);
					lastSweepNodeElement = actualSweepNodeElement; lastEventKey = actualEventKey;
				}
			}
			else if (actualEventKey.type == TwoPolygonsEventStruct.CROSS)
			{
				// Upper e lower so trocados com relao a antecessor e successor pois essa
				// funo  chamada depois da ordem deles ter sido invertida. Representam as
				// arestas acima e abaixo do cruzamento ANTES do cruzamento.
				TwoPolygonsSweepLine.SweepLineElement lowerElement = (TwoPolygonsSweepLine.SweepLineElement) ((TwoPolygonsEventStruct.EventCrossKey) actualEvent.key).antecessorNode.element;
				TwoPolygonsSweepLine.SweepLineElement upperElement = (TwoPolygonsSweepLine.SweepLineElement) ((TwoPolygonsEventStruct.EventCrossKey) actualEvent.key).successorNode.element;

				Vertex vNew = new Vertex(actualEventKey.pEvt);
				if (lowerElement.getVertex() != null)
				{
					if (upperElement.getVertex() != null)
					{
						vNew.vBegin = lowerElement.getVertex().vBegin;
						vNew.next = lowerElement.getVertex();
						vNew.previous = upperElement.getVertex();
						upperElement.getVertex().next = vNew;
						lowerElement.getVertex().previous = vNew;

						upperElement.setVertex(null);
						lowerElement.setVertex(null);
					}
					else
					{
						vNew.vBegin = lowerElement.getVertex().vBegin;
						if (lowerElement.getUpperChain())
						{
							vNew.next = lowerElement.getVertex();
							lowerElement.getVertex().previous = vNew;
						}
						else
						{
							vNew.previous = lowerElement.getVertex();
							lowerElement.getVertex().next = vNew;
						}
						upperElement.setVertex(vNew);
						upperElement.setUpperChain(lowerElement.getUpperChain());
						lowerElement.setVertex(null);
					}
				}
				else 
				{
					if (upperElement.getVertex() != null)
					{
						vNew.vBegin = upperElement.getVertex().vBegin;
						if (upperElement.getUpperChain())
						{
							vNew.next = upperElement.getVertex();
							upperElement.getVertex().previous = vNew;
						}
						else
						{
							vNew.previous = upperElement.getVertex();
							upperElement.getVertex().next = vNew;
						}
						lowerElement.setVertex(vNew);
						lowerElement.setUpperChain(upperElement.getUpperChain());
						upperElement.setVertex(null);
					}
					
					// Com nenhuma das arestas associadas, transforma uma regio AB em uma regio VACUUM.
					else
					{
						Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
						vNewEntry.vBegin = vNewEntry;
						put(vNewEntry, null);
						upperElement.setVertex(vNewEntry); upperElement.setUpperChain(true);
						lowerElement.setVertex(vNewEntry); lowerElement.setUpperChain(false);
					}
				}
			}
		}
		else if (selectedType == TwoPolygonsOutputGenerator.INTERSECTION)
		{
			if (actualEventKey.type == TwoPolygonsEventStruct.BEGIN)
			{
				TwoPolygonsSweepLine.SweepLineElement actualSweepNodeElement = null;
				actualSweepNodeElement = (TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element;

				// Sendo esse ponto-evento no-coincidente com o passado ...
				if (lastEventKey == null || !lastEventKey.pEvt.equals(actualEventKey.pEvt))
				{
					int actualLeftOwner = ((TwoPolygonsRegionsMarkersTerminatorFactory.Element) actualEventKey.sweepNode.getLeftTerminator().element).getOwner();
					int owner = ((TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element).getOwner();
					int notOwner = (owner == TwoPolygonsCrossings.A) ? TwoPolygonsCrossings.B : TwoPolygonsCrossings.A;
					// Se a aresta estiver sobre uma regio notOwner ...
					if (actualLeftOwner == notOwner)
					{
						Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
						vNewEntry.vBegin = vNewEntry;
						put(vNewEntry, null);
						actualSweepNodeElement.setVertex(vNewEntry);
						actualSweepNodeElement.setUpperChain(false);
						lastSweepNodeElement = actualSweepNodeElement; lastEventKey = actualEventKey;
					}
					// Se a aresta estiver sobre uma regio owner + notOwner ...
					else if (actualLeftOwner == owner + notOwner)
					{
						Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
						vNewEntry.vBegin = vNewEntry;
						put(vNewEntry, null);
						actualSweepNodeElement.setVertex(vNewEntry);
						actualSweepNodeElement.setUpperChain(true);
						lastSweepNodeElement = actualSweepNodeElement; lastEventKey = actualEventKey;
					}
				}

				// Tendo BEGIN igual o mesmo ponto-evento que o anterior, pode ser duas coisas:
				else if (lastEventKey != null && lastEventKey.pEvt.equals(actualEventKey.pEvt))
				{
					// Pode ser a parte de cima de um outro BEGIN ...
					if (actualEventKey.type == lastEventKey.type)
					{
						int actualLeftOwner = ((TwoPolygonsRegionsMarkersTerminatorFactory.Element) actualEventKey.sweepNode.getLeftTerminator().element).getOwner();
						int owner = ((TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element).getOwner();
						int notOwner = (owner == TwoPolygonsCrossings.A) ? TwoPolygonsCrossings.B : TwoPolygonsCrossings.A;
						// ... e deve delimitar uma regio apropriada para INTERSECTION.
						if (actualLeftOwner == owner + notOwner || actualLeftOwner == notOwner)
						{
							boolean lastLeftUpperChain = ((TwoPolygonsSweepLine.SweepLineElement) lastEventKey.sweepNode.element).getUpperChain();
							actualSweepNodeElement.setVertex(lastSweepNodeElement.getVertex());
							actualSweepNodeElement.setUpperChain(!lastLeftUpperChain);
						}
					}

					// ... ou pode ser a continuao de um END. Quem cria o vertex  o evento de END,
					// no o de BEGIN. Ele s associa o vrtice.
					else
					{
						actualSweepNodeElement.setVertex(lastSweepNodeElement.getVertex());
						actualSweepNodeElement.setUpperChain(lastSweepNodeElement.getUpperChain());
					}
				}
			}
			else if (actualEventKey.type == TwoPolygonsEventStruct.END)
			{
				TwoPolygonsSweepLine.SweepLineElement actualSweepNodeElement = (TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element;

				// Se o ponto-evento anterior for igual ao atual e se pertencerem ao mesmo polgono,
				//  ento quer dizer que foi um evento de END fechando as cadeias, e essa ltima 
				//  a superior.
				if (lastEventKey != null && 
				    lastEventKey.pEvt.equals(actualEventKey.pEvt) && 
				    ((TwoPolygonsSweepLine.SweepLineElement) lastEventKey.sweepNode.element).getOwner() == 
				    ((TwoPolygonsSweepLine.SweepLineElement) actualEventKey.sweepNode.element).getOwner())
				{
					if (actualSweepNodeElement.getUpperChain())
					{
						lastSweepNodeElement.getVertex().next = actualSweepNodeElement.getVertex();
						actualSweepNodeElement.getVertex().previous = lastSweepNodeElement.getVertex();
					}
					else
					{
						lastSweepNodeElement.getVertex().previous = actualSweepNodeElement.getVertex();
						actualSweepNodeElement.getVertex().next = lastSweepNodeElement.getVertex();
					}
					lastEventKey = null;
				}

				// Se o ponto-evento no for o mesmo mas o sweepNode.Element estiver associado, ento 
				// um novo vrtice a ser adicionado.
				else if (actualSweepNodeElement.getVertex() != null)
				{
					Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
					vNewEntry.vBegin = actualSweepNodeElement.getVertex().vBegin;

					// Se  pra adicionar na cadeia superior ...
					if (actualSweepNodeElement.getUpperChain())
					{
						actualSweepNodeElement.getVertex().previous = vNewEntry;
						vNewEntry.next = actualSweepNodeElement.getVertex();
					}

					// ... ou na cadeia inferior.
					else
					{
						actualSweepNodeElement.getVertex().next = vNewEntry;
						vNewEntry.previous = actualSweepNodeElement.getVertex();
					}
					actualSweepNodeElement.setVertex(vNewEntry);
					lastSweepNodeElement = actualSweepNodeElement; lastEventKey = actualEventKey;
				}
			}
			else if (actualEventKey.type == TwoPolygonsEventStruct.CROSS)
			{
				// Upper e lower so trocados com relao a antecessor e successor pois essa
				// funo  chamada depois da ordem deles ter sido invertida. Representam as
				// arestas acima e abaixo do cruzamento ANTES do cruzamento.
				TwoPolygonsSweepLine.SweepLineElement lowerElement = (TwoPolygonsSweepLine.SweepLineElement) ((TwoPolygonsEventStruct.EventCrossKey) actualEvent.key).antecessorNode.element;
				TwoPolygonsSweepLine.SweepLineElement upperElement = (TwoPolygonsSweepLine.SweepLineElement) ((TwoPolygonsEventStruct.EventCrossKey) actualEvent.key).successorNode.element;

				Vertex vNew = new Vertex(actualEventKey.pEvt);
				if (lowerElement.getVertex() != null)
				{
					if (upperElement.getVertex() != null)
					{
						vNew.vBegin = lowerElement.getVertex().vBegin;
						vNew.next = upperElement.getVertex();
						vNew.previous = lowerElement.getVertex();
						lowerElement.getVertex().next = vNew;
						upperElement.getVertex().previous = vNew;

						upperElement.setVertex(null);
						lowerElement.setVertex(null);
					}
					else
					{
						vNew.vBegin = lowerElement.getVertex().vBegin;
						if (lowerElement.getUpperChain())
						{
							vNew.next = lowerElement.getVertex();
							lowerElement.getVertex().previous = vNew;
						}
						else
						{
							vNew.previous = lowerElement.getVertex();
							lowerElement.getVertex().next = vNew;
						}
						upperElement.setVertex(vNew);
						upperElement.setUpperChain(lowerElement.getUpperChain());
						lowerElement.setVertex(null);
					}
				}
				else 
				{
					if (upperElement.getVertex() != null)
					{
						vNew.vBegin = upperElement.getVertex().vBegin;
						if (upperElement.getUpperChain())
						{
							vNew.next = upperElement.getVertex();
							upperElement.getVertex().previous = vNew;
						}
						else
						{
							vNew.previous = upperElement.getVertex();
							upperElement.getVertex().next = vNew;
						}
						lowerElement.setVertex(vNew);
						lowerElement.setUpperChain(upperElement.getUpperChain());
						upperElement.setVertex(null);
					}
					
					// Com nenhuma das arestas associadas, transforma uma regio AB em uma regio VACUUM.
					else
					{
						Vertex vNewEntry = new Vertex(actualEventKey.pEvt);
						vNewEntry.vBegin = vNewEntry;
						put(vNewEntry, null);
						upperElement.setVertex(vNewEntry); upperElement.setUpperChain(false);
						lowerElement.setVertex(vNewEntry); lowerElement.setUpperChain(true);
					}
				}
			}
		}
	}


	/**
	 * Retorna a estrutura contendo a informao de sada.
	 */
	public Object getAnswer()
	{
		if (selectedType == TwoPolygonsOutputGenerator.CROSS_POINTS)
		{
			List listRetVal = new LinkedList();
			AVLTerminatedTree.AVLNode nodeTmp = getLeftMostNodeAndRemoveIt();
			while (nodeTmp != null)
			{
				listRetVal.add(((Vertex) nodeTmp.key).ptVertex);
				nodeTmp = getLeftMostNodeAndRemoveIt();
			}
			return listRetVal;
		}
		else if (selectedType == TwoPolygonsOutputGenerator.UNION || selectedType == TwoPolygonsOutputGenerator.INTERSECTION)
		{
			List listPolygon = null, listRetVal = new LinkedList();
			AVLTerminatedTree.AVLNode nodeTmp = getLeftMostNodeAndRemoveIt();
			Vertex vertexFirst = null, vertexIt = null;
			while (nodeTmp != null)
			{
				listPolygon = new LinkedList();
				vertexFirst = (Vertex) nodeTmp.key;
				listPolygon.add(vertexFirst.ptVertex);
				vertexIt = vertexFirst.next;
				while (vertexIt != vertexFirst)
				{
					listPolygon.add(vertexIt.ptVertex);
					if (vertexIt == vertexIt.vBegin)
						remove(vertexIt);
					vertexIt = vertexIt.next;
				}
				listRetVal.add(listPolygon);
				nodeTmp = getLeftMostNodeAndRemoveIt();
			}
			return listRetVal;
		}
		else if (selectedType == TwoPolygonsOutputGenerator.INTERSECTION+1000)
		{
			List listRetVal = new LinkedList();
			List listPolygon = null;
			AVLTerminatedTree.AVLNode nodeTmp = getLeftMostNodeAndRemoveIt();
			Vertex vertexFirst = null, vertexIt = null;
			while (nodeTmp != null)
			{
				listPolygon = new LinkedList();
				vertexFirst = (Vertex) nodeTmp.key;
				listPolygon.add(vertexFirst.ptVertex);
				vertexIt = vertexFirst.next;
				while (vertexIt != vertexFirst)
				{
					listPolygon.add(vertexIt.ptVertex);
					if (vertexIt == vertexIt.vBegin)
						remove(vertexIt);
					vertexIt = vertexIt.next;
				}
				listRetVal.add(listPolygon);
			}
			return listRetVal;
		}
		else
			return null;
	}

	/**
	 * Classe usada para armazenar as posies dos vrtices na esturtura que gera a
	 * informao de sada. Em cada objeto Vertex est a ifnormao sobre um ponto,
	 * sobre os elementos adjacentes a ele na estrutura e se ele est contido na
 	 * na rvore principal ou s nas listas geradas a partir de um elemento na rvore
 	 * principal.
	 */
	public static class Vertex
	{
		// Guarda o valor desse vrtice.
		Point2D ptVertex = null;

		// Guarda o vrtice da cadeia que faz parte que est inserido na estrutura.
		Vertex vBegin = null;

		// next e previous da lista de vrtices.
		Vertex next = null;
		Vertex previous = null;

		/**
		 * Cria um novo objeto.
		 * @param   pEvt	o ponto que representa o vrtice sendo armazenado.
		 */
		public Vertex(Point2D pEvt)	{ptVertex = pEvt;}


		public String toString()
		{
			String ret = "ptVertex:(" + ptVertex.getX() + ", " + ptVertex.getY() + ")";
			if (previous != null)
				ret += " previous.ptVertex:(" + previous.ptVertex.getX() + ", " + previous.ptVertex.getY() + ")";
			if (next != null)
				ret += " next.ptVertex:(" + next.ptVertex.getX() + ", " + next.ptVertex.getY() + ")";
			return ret;
		}
	}


	/**
	 * Classe que determina se um n da estrututura TwoPolygonsOutputGenerator  maior, menor ou
	 * igual a outro n.
	 */
	public static class OutputComparator implements Comparator
	{
		/**
		 * Mtodo que realiza a comparao entre o1 e o2, onde o1 e o2 so ns da estrutura em questo.
		 * O critrio  o seguinte: o1  maior que o2 se:
		 *	1- O pEvt da chave de  o1  maior com relao a x-coordenada do que pEvt da chave de  o2.
		 *	2- O pEvt da chave de  o1  maior com relao a y-coordenada do que pEvt da chave de  o2.
		 * @param   o1	uma das chaves a ser comparadas. 
		 * 		o2	outra das chaves a ser comparadas. 
		 */
		public int compare(Object o1, Object o2)
		{
			Vertex key1 = (Vertex) o1;
			Vertex key2 = (Vertex) o2;
			return GeomOperations.compareByY(key1.ptVertex, key2.ptVertex);
		}
	}


	/**
	 * Classe que determina o que fazer caso dois forem dados como iguais pelo comparator.
	 * Nesse caso no faz nada e mantm a deciso do comparator.
	 */
	public static class OutputResolutor implements Resolutor
	{		
		public int resolve(INodeMap nodeOne, INodeMap nodeTwo)	{return 0;}
		public int resolveKeys(Object keyOne, Object keyTwo)		{return 0;}
	}


	/**
	 * Mtodo para debug. Imprime o valor dos terminators da estrutura.
	 */
	public void debugTerminators()
	{
		System.out.println("\noutPutGenerator.debugTerminators");	
		if (head.left == null || head.left instanceof AVLTerminatedTree.Terminator) return;
		AVLTerminatedTree.AVLNode thisNode = head.left;
		Vertex firstVertex = (Vertex) thisNode.key, vIt = firstVertex.next;
		System.out.print(" next: | firstVertex.ptVertex:" + firstVertex.ptVertex.getX() + ", " + firstVertex.ptVertex.getY() + ")");	
		while (vIt != null && vIt != firstVertex)
		{
			System.out.print(" | vIt.ptVertex:" + vIt.ptVertex.getX() + ", " + vIt.ptVertex.getY() + ")");
			vIt = vIt.next;
		}
		System.out.println("");
		if (vIt != null)
		{
			System.out.println("vIt != null");
			if (vIt != vIt.vBegin)
				System.out.println("vIt != vIt.vBegin");
			if (vIt != firstVertex)
				System.out.println("vIt != firstVertex");
		}
		else
			System.out.println("vIt == null");

		vIt = firstVertex.previous;
		System.out.print(" previous: | firstVertex.ptVertex:" + firstVertex.ptVertex.getX() + ", " + firstVertex.ptVertex.getY() + ")");	
		while (vIt != null && vIt != firstVertex)
		{
			System.out.print(" | vIt.ptVertex:" + vIt.ptVertex.getX() + ", " + vIt.ptVertex.getY() + ")");
			vIt = vIt.previous;
		}
		System.out.println("");
		if (vIt != null)
		{
			System.out.println("vIt != null");
			if (vIt != vIt.vBegin)
				System.out.println("vIt != vIt.vBegin");
			if (vIt != firstVertex)
				System.out.println("vIt != firstVertex");
		}
		else
			System.out.println("vIt == null");
	}
}