/*
 * @(#)TwoPolygonsRegionsMarkersTerminatorFactory.java	1.0 2000/07/22
 *
 */
package rob.math.geom;

import java.awt.Color;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.util.Collection;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Comparator;
import rob.math.*;
import rob.util.*;
import rob.exceptions.*;


/**
 * Um objeto <code>TwoPolygonsRegionsMarkersTerminatorFactory</code> produz Terminators para uma
 * <code>AVLTerminatedTree</code>. Esse tipode rvore no acaba suas folhas apontando para null, mas
 * sim para um Terminator, e nessa rvore em especial,  garantido quesempre que um terminator  
 * removido ou adicionado,  feito de tal modo que as duas operaes ocorram com temrinators apontados
 * por um link right de uma folha.
 * Os terminators produzidos por essa classe conteram um identificador que representa a regio entre dois
 * ns adjecentes na rvore, que representam divises de regies. Por exemplo, se na rvore temos dois ns
 * adjacentes qeu representam arestas do polgono A, ento entre eles na rvore existir ou um Terminator 
 * marcado com A ou A+B, e logo aps cada um deles um Terminator marcado com vazio (VACUUM) ou com B.
 * Essa classe  encarregada de manter a integridade da informao armazenada nos Terminators..
 * <p>
 *
 * @version 	1.0 2000/07/22
 * @author 	Roberto Giordano Barra
 * @see         
 */
public class TwoPolygonsRegionsMarkersTerminatorFactory extends AVLRightTerminatedTree.TerminatorFactory
{
	/**
	 * Cria um terminator do tipo BEGIN baseado baseado na folha passada.
	 * @param   nodeForMe	a folha que ser pai do terminator e que o tem associado como filho
	 *				direito.
	 * @returns   o terminator criado. 
	 */
	public AVLRightTerminatedTree.Terminator createTerminator(AVLRightTerminatedTree.AVLNode nodeForMe, boolean forLeft)
	{
		//  Tabela que obedece:
		//  left Terminator | this Terminator
		//  ---------------------------------
		//   null           |  vacuum
		//   vacuum         |  owner
		//   owner          |  vacuum
		//                 |   + owner
		//    + owner      |  
		//                  |
		//                  |

		AVLRightTerminatedTree.RightTerminator retTerm = new AVLRightTerminatedTree.RightTerminator();
		retTerm.parent = nodeForMe;

		int owner = ((TwoPolygonsSweepLine.SweepLineElement) nodeForMe.getElement()).getOwner();
		int notOwner = (owner == TwoPolygonsCrossings.A) ? TwoPolygonsCrossings.B : TwoPolygonsCrossings.A;

		//  o primeiro ou a ser adicionado.
		if (nodeForMe.left == null)
			retTerm.element = new Element(TwoPolygonsCrossings.VACUUM);
		else 
		{
			// Sempre  criado um terminator para o brao direito de uma folha, ou seja,
			// o brao esquerdo sempre ter um terminator (a hiptese de ser null j foi
			// eliminada.
			AVLRightTerminatedTree.Terminator leftTerminator = (AVLRightTerminatedTree.Terminator) nodeForMe.left;
			int leftRegionValue = ((Element) leftTerminator.element).getOwner();
			if (leftRegionValue == TwoPolygonsCrossings.VACUUM)
				retTerm.element = new Element(owner);
			else if (leftRegionValue == owner)
				retTerm.element = new Element(TwoPolygonsCrossings.VACUUM);
			else if (leftRegionValue == notOwner)
				retTerm.element = new Element(owner + notOwner);
			else if (leftRegionValue == owner + notOwner)
				retTerm.element = new Element(notOwner);
			addNewElement(retTerm, leftTerminator);
		}
		return retTerm;
	}


	public void addNewElement(AVLRightTerminatedTree.AVLNode nodeRetTerm, AVLRightTerminatedTree.AVLNode leftTerminator)
	{
		nodeRetTerm.right = leftTerminator.right;
		if (leftTerminator.right != null)
			leftTerminator.right.left = leftTerminator;
		nodeRetTerm.left = leftTerminator;
		leftTerminator.right = nodeRetTerm;
	}


	/**
	 * Atualiza a estrutura com relao ao ponto-evento passado. S vo interessar os ponto-eventos de 
	 * interseco. A remoo dos terminators inteis  feita automaticamente quando se desloca pela
 	 * lista formada pelos terminators se for utilizado os mtodos next() e previous().
	 * @param   actualEvent	o ponto-evento correspondente. 
	 */
	public void stepTerminator(AVLTerminatedTree.AVLNode actualEvent)
	{
		TwoPolygonsEventStruct.EventStructKey actualEventKey = (TwoPolygonsEventStruct.EventStructKey) actualEvent.key;
		if (((TwoPolygonsEventStruct.EventStructKey) actualEvent.key).type == TwoPolygonsEventStruct.CROSS)
		{
			//  before CROSS 	  | after CROSS
			//  ---------------------------------
			//   VACUUM         |   + owner
			//   owner          |  
			//                 |  owner
			//    + owner      |  VACUUM
			//                  |

			TwoPolygonsEventStruct.EventCrossKey actualEventCrossKey = (TwoPolygonsEventStruct.EventCrossKey) actualEvent.key;
			AVLRightTerminatedTree.AVLNode upperNode = actualEventCrossKey.successorNode;

			// Pega terminator com regio entre antecessorNode e successorNode.
			Element elemTmp = (Element) ((AVLRightTerminatedTree.Terminator) upperNode.getLeftTerminator()).element;
			if (elemTmp.getOwner() == TwoPolygonsCrossings.VACUUM)
				elemTmp.setOwner(TwoPolygonsCrossings.A + TwoPolygonsCrossings.B);
			else if (elemTmp.getOwner() == TwoPolygonsCrossings.A + TwoPolygonsCrossings.B)
				elemTmp.setOwner(TwoPolygonsCrossings.VACUUM);
			else if (elemTmp.getOwner() == TwoPolygonsCrossings.A)
				elemTmp.setOwner(TwoPolygonsCrossings.B);
			else if (elemTmp.getOwner() == TwoPolygonsCrossings.B)
				elemTmp.setOwner(TwoPolygonsCrossings.A);
		}
	}


	/**
	 * Classe que representa o elemento de um terminator.
	 */
	public static class Element
	{
		// Armazena a regio que o terminator representa.
		private int owner;

		public Element(int ownerNew)	{owner = ownerNew;}


		public int getOwner()	{return owner;}
		public void setOwner(int newOwner)	{owner = newOwner;}
	}
}