package rob.util;


import java.util.Comparator;


public class AVLTerminatedTree
{
	// Objeto usado para comparar elementos enquanto so
	// inseridos na rvore. O elemento a ser inserido
	// sempre ser o primeiro parmetro passado para o
	// mtodo compare().
	protected Comparator comp;

	// Objeto usado para decidir o que fazer quando duas
	// chaves consideradas iguais so encontradas. O mtodo
	// resolve(Node element1, Node element2)  ento
	// chamado e em caso de retorno zero a execuo da 
	// funco para de descer na rvore e parte para a
	// finalizao do mtodo. No esquecer que o Resolutor
	// deve possuir um apontador para a rvore em questo
	protected Resolutor resol;
		

	protected TerminatorFactory termFactory;


	// O n dessa rvore tem parent apontando para s mesmo
	// e left apontando para o primeiro elemento da rvore.
	protected AVLNode head;


	protected long size = 0;


	public AVLTerminatedTree(Comparator compNew, Resolutor resolNew, TerminatorFactory termFactoryNew)
	{
		head = new AVLNode();
		comp = compNew; 
		resol = resolNew;
		termFactory = termFactoryNew;
	}


	public static final byte PRE_ORDER = 1;
	public static final byte IN_ORDER = 2;
	public static final byte POS_ORDER = 3;
	public void traverse(byte mode)
	{
		if (head.left == null)
			return;
		if (mode == PRE_ORDER)
			traversePre(head.left);
		else if (mode == POS_ORDER)
			traversePos(head.left);
		else // if (mode == IN_ORDER)
			traverseIn(head.left);
	}


	protected void traversePre(AVLNode theOne)
	{
		System.out.print("|" + theOne.key.toString() + ", " + theOne.balance + "|");
		if (!(theOne.left instanceof Terminator))
			traversePre(theOne.left);
		if (!(theOne.right instanceof Terminator))
			traversePre(theOne.right);
	}


	protected void traversePos(AVLNode theOne)
	{
		if (!(theOne.left  instanceof Terminator))
			traversePos(theOne.left);
		if (!(theOne.right  instanceof Terminator))
			traversePos(theOne.right);
		System.out.print("|" + theOne.key.toString() + ", " + theOne.balance + "|");
	}


	protected void traverseIn(AVLNode theOne)
	{
		if (!(theOne.left  instanceof Terminator))
			traverseIn(theOne.left);
		System.out.print("|" + theOne.key.toString() + ", " + theOne.balance + "|");
		if (!(theOne.right  instanceof Terminator))
			traverseIn(theOne.right);
	}


	public void put(Object key, Object element)
	{
		addNode(new AVLNode(key, element));
	}


	protected void addNode(AVLNode nodeNew)
	{
		if (head.left == null)
		{
			size++;
			head.left = nodeNew;
			head.left.parent = head;
			head.left.left = termFactory.createTerminator(head.left, true);
			head.left.right = termFactory.createTerminator(head.left, false);
			return;
		}

		AVLNode nodeIterator = head.left;
		int result;
		while (true)
		{
			result = comp.compare(nodeNew.key, nodeIterator.key);

			// Decide o que fazer com o n em caso de valores de chaves iguais.
			if (result == 0)
				result = resol.resolve(nodeNew, nodeIterator);

			// Adiciona na esquerda.
			if  (result  < 0)
			{
				if  (nodeIterator.left instanceof Terminator)
				{
					nodeNew.parent = nodeIterator;
					nodeNew.right = nodeIterator.left;
					nodeNew.right.parent = nodeNew;
					nodeIterator.left = nodeNew;
					nodeNew.left = termFactory.createTerminator(nodeNew, true);
					fixupInsertion(nodeNew);
					return;
				}
				else
				{
					nodeIterator = nodeIterator.left;
					continue;
				}
			}

			// Adiciona na direita.
			else if (result  > 0)
			{
				if  (nodeIterator.right  instanceof Terminator)
				{
					nodeNew.parent = nodeIterator;
					nodeNew.left = nodeIterator.right;
					nodeNew.left.parent = nodeNew;
					nodeIterator.right = nodeNew;
					nodeNew.right = termFactory.createTerminator(nodeNew, false);
					fixupInsertion(nodeNew);
					return;
				}
				else
				{
					nodeIterator = nodeIterator.right;
					continue;
				}
			}
			// Adiciona no mesmo.
			else
			{
				nodeNew.parent = nodeIterator.parent;
				nodeNew.left = nodeIterator.left;
				nodeNew.right = nodeIterator.right;
				nodeNew.balance = nodeIterator.balance;
				if (nodeIterator.parent.left == nodeIterator)
					nodeIterator.parent.left = nodeNew;
				else
					nodeIterator.parent.right = nodeNew;
				nodeIterator.right.parent = nodeNew;
				nodeIterator.left.parent = nodeNew;
				return;
			}
		}
	}


	public Object get(Object key)
	{
		if (head.left == null)
			return null;
		AVLNode nodeIterator = head.left;
		int result = 0;
		while (!(nodeIterator instanceof Terminator))
		{
			result = comp.compare(key, nodeIterator.key);

			// Decide o que fazer com o n em caso de valores de chaves iguais.
			if (result == 0)
				result = resol.resolveKeys(key, nodeIterator.key);

 			// Desce para esquerda.
			if (result > 0)
				nodeIterator = nodeIterator.right;

			// Desce para direita.
			else if  (result < 0)
				nodeIterator = nodeIterator.left;

			// Achou!
			else
				return nodeIterator.element;
		}
		return null;
	}


	public Object remove(Object key)
	{
		if (head.left == null)
			return null;
		AVLNode nodeIterator = head.left;
		int result = 0;
		while (!(nodeIterator instanceof Terminator))
		{
			result = comp.compare(key, nodeIterator.key);
			if (result == 0)
				result = resol.resolveKeys(key, nodeIterator.key);
			if  (result > 0)
				nodeIterator = nodeIterator.right;
			else if  (result < 0)
				nodeIterator = nodeIterator.left;
			// Achou!
			else
			{
				deleteNode(nodeIterator);
				return nodeIterator.element;
			}
		}
		return null;
	}


/**
 * 	Remove um n X.
 */
	protected void deleteNode(AVLNode nodeIterator)
	{	
		if (nodeIterator.left instanceof Terminator)
		{
			// n no tem filhos.
			if (nodeIterator.right instanceof Terminator)
			{
System.out.print("<n no tem filhos.>");
				if (nodeIterator.parent.left == nodeIterator)
				{
					nodeIterator.parent.left = nodeIterator.right;
					nodeIterator.right.parent = nodeIterator.parent;
					nodeIterator.left.parent = null;
					fixupRemotion(nodeIterator.parent, 1);
				}
				else
				{
					nodeIterator.parent.right = nodeIterator.left;
					nodeIterator.left.parent = nodeIterator.parent;
					nodeIterator.right.parent = null;
					fixupRemotion(nodeIterator.parent, -1);
				}
			}
			// s tem filho direito.
			else
			{
System.out.print("<s tem filho direito.>");
				nodeIterator.left.parent = null;
				nodeIterator.right.parent = nodeIterator.parent;
				if (nodeIterator.parent.left == nodeIterator)
				{
					nodeIterator.parent.left = nodeIterator.right;
					fixupRemotion(nodeIterator.parent, 1);
				}
				else
				{
					nodeIterator.parent.right = nodeIterator.right;
					fixupRemotion(nodeIterator.parent, -1);
				}
			}
		}
		else
		{
			// s tem filho esquerdo.
			if (nodeIterator.right instanceof Terminator)
			{
System.out.print("<s tem filho esquerdo.>");
				nodeIterator.right.parent = null;
				nodeIterator.left.parent = nodeIterator.parent;
				if (nodeIterator.parent.left == nodeIterator)
				{
					nodeIterator.parent.left = nodeIterator.left;
					fixupRemotion(nodeIterator.parent, 1);
				}
				else
				{
					nodeIterator.parent.right = nodeIterator.left;
					fixupRemotion(nodeIterator.parent, -1);
				}
			}
			// tem os dois filhos.
			else
			{
System.out.print("<tem os dois filhos: nodeIterator="+nodeIterator+" nodeIterator.right="+nodeIterator.right);
System.out.print(" nodeIterator.right.left"+nodeIterator.right.left+">");
				if (nodeIterator.right.left instanceof Terminator)
				{
System.out.print("<nodeIterator.right.left instanceof Terminator>");
					nodeIterator.right.left.parent = null;
					if (nodeIterator.parent.left == nodeIterator)
						nodeIterator.parent.left = nodeIterator.right;
					else
						nodeIterator.parent.right = nodeIterator.right;
					nodeIterator.right.left = nodeIterator.left;
					nodeIterator.left.parent = nodeIterator.right;
					nodeIterator.right.parent = nodeIterator.parent;
					nodeIterator.right.balance = nodeIterator.balance;
					fixupRemotion(nodeIterator.right, -1);
				}
				else
				{
System.out.print("<!(nodeIterator.right.left instanceof Terminator)");
					AVLNode changeNode = nodeIterator.right;
					while (!(changeNode.left instanceof Terminator))
						changeNode = changeNode.left;
					changeNode.left.parent = null;
					changeNode.parent.left = changeNode.right;
					changeNode.right.parent = changeNode.parent;
					AVLNode nodeTmp = changeNode.parent;
					changeNode.parent = nodeIterator.parent;
					changeNode.right = nodeIterator.right;
					changeNode.left = nodeIterator.left;
					changeNode.balance = nodeIterator.balance;
					if (nodeIterator.parent.left == nodeIterator)
						nodeIterator.parent.left = changeNode;
					else
						nodeIterator.parent.right = changeNode;
					nodeIterator.right.parent = changeNode;
					nodeIterator.left.parent = changeNode;
					fixupRemotion(nodeTmp, 1);
				}
			}
		}
		if (head.left instanceof Terminator)
			head.left = null;
	}


/**
 * 	D um jeito na remoo. diff diz a diferena de altura atual.
 */
	protected void fixupRemotion(AVLNode toBeChanged, int diff)
	{
		size--;
		if (toBeChanged == head)
			return;

		// Para o primeiro n, a rvore vai seguir a mudana de altura.
		toBeChanged.balance += diff;
		if (toBeChanged.balance == 2)
		{
			if (toBeChanged.right.balance == 0)
			{
				shrinkRightSubTree(toBeChanged);
				return;
			}
			else if (toBeChanged.right.balance > 0)
				shrinkRightSubTree(toBeChanged);
			else
			{
				shrinkLeftSubTree(toBeChanged.right);
				shrinkRightSubTree(toBeChanged);
			}
				toBeChanged = toBeChanged.parent;
		}
		else if (toBeChanged.balance == -2)
		{
			if (toBeChanged.left.balance == 0)
			{
				shrinkLeftSubTree(toBeChanged);
				return;
			}
			else if (toBeChanged.left.balance < 0)
				shrinkLeftSubTree(toBeChanged);
			else
			{
				shrinkRightSubTree(toBeChanged.left);
				shrinkLeftSubTree(toBeChanged);
			}
				toBeChanged = toBeChanged.parent;
		}
		else
		{
			if (toBeChanged.balance != 0)
				return;
		}

		AVLNode nodeParent = toBeChanged.parent;
		while (nodeParent != head)
	  	{
			if (nodeParent.left == toBeChanged)
			{
				if (nodeParent.balance == 0)
				{
					nodeParent.balance = 1;
					return;
				}
				else if (nodeParent.balance == -1)
					nodeParent.balance = 0;
				// Rotao.
				else
				{
					nodeParent.balance = 2;
					if (nodeParent.right.balance == 0)
					{
						shrinkRightSubTree(nodeParent);
						return;
					}
					else if (nodeParent.right.balance > 0)
						shrinkRightSubTree(nodeParent);
					else
					{
						shrinkLeftSubTree(nodeParent.right);
						shrinkRightSubTree(nodeParent);
					}
					nodeParent = nodeParent.parent;
				}
			}
			else
			{
				if (nodeParent.balance == -1)
				{
					nodeParent.balance = -2;
					if (nodeParent.left.balance == 0)
					{
						shrinkLeftSubTree(nodeParent);
						return;
					}
					else if (nodeParent.left.balance < 0)
						shrinkLeftSubTree(nodeParent);
					else
					{
						shrinkRightSubTree(nodeParent.left);
						shrinkLeftSubTree(nodeParent);
					}
					nodeParent = nodeParent.parent;
				}
				else if (nodeParent.balance == 1)
					nodeParent.balance = 0;
				else
				{
					nodeParent.balance = -1;
					return;
				}
			}
			toBeChanged = nodeParent;
			nodeParent = nodeParent.parent;
		}
	}


/**
 * 	D um jeito na insero.
 */
	public void fixupInsertion(AVLNode nodeNew)
	{
		size++;
		AVLNode nodeParent = nodeNew.parent;
		while (nodeParent != head)
	  	{
			if (nodeParent.left == nodeNew)
			{
				// Se a rvore da direita j era maior do que a da esquerda, no muda de tamanho, s diminue o balano.
				if (nodeParent.balance > 0)
				{
					nodeParent.balance = 0;
					return;
				}

				// Aqui a rvore da direita  igual a da esquerda, ento muda de tamanho e desbalanceia.
				else if (nodeParent.balance == 0)
					nodeParent.balance = -1;

				// Rotao.
				else
				{
					nodeParent.balance = -2;
					if  (nodeParent.left.balance == -1)
						shrinkLeftSubTree(nodeParent);
					else
					{
						shrinkRightSubTree(nodeParent.left);
						shrinkLeftSubTree(nodeParent);
					}
					nodeParent = nodeParent.parent;
					return ;
				}
			}
			else
			{
				// Se a rvore da direita j era maior do que a da esquerda, ento aumenta a rvore e a rotaciona.
				if (nodeParent.balance > 0)
				{
					nodeParent.balance = 2;
					if  (nodeParent.right.balance == 1)
						shrinkRightSubTree(nodeParent);
					else
					{
						shrinkLeftSubTree(nodeParent.right);
						shrinkRightSubTree(nodeParent);
					}
					nodeParent = nodeParent.parent;
					return ;
				}

				// Aqui a rvore da direita  igual a da esquerda, ento muda de tamanho e desbalanceia.
				else if (nodeParent.balance == 0)
					nodeParent.balance = 1;

				// Zera o balance e para o algoritmo.
				else
				{
					nodeParent.balance = 0;
					return ;
				}

			}
			// Caminha em direo do prximo do topo da rvore.
			nodeNew = nodeParent;
			nodeParent = nodeParent.parent;
		}
	}


	public void shrinkLeftSubTree(AVLNode toBeLowered)
	{
		// Ajusta o balano dos ns.
		if (toBeLowered.left.balance < 0)
			toBeLowered.balance = toBeLowered.balance + 1 - toBeLowered.left.balance;
		else
			toBeLowered.balance++;;
		if (toBeLowered.balance > 0)
			toBeLowered.left.balance = toBeLowered.left.balance + 1 + toBeLowered.balance;
		else
			toBeLowered.left.balance++;

		// Cuida de arrumar o pai de todos (no t na jogada).
		if  (toBeLowered.parent.left == toBeLowered)
			toBeLowered.parent.left = toBeLowered.left;
		else
			toBeLowered.parent.right = toBeLowered.left;

		// Cuida das trocas.
		toBeLowered.left.parent = toBeLowered.parent; 
		toBeLowered.parent = toBeLowered.left; 
		toBeLowered.left  = toBeLowered.left.right;
		toBeLowered.left.parent = toBeLowered;
		toBeLowered.parent.right = toBeLowered;
	}


	public void shrinkRightSubTree(AVLNode toBeLowered)
	{
		// Ajusta o balano dos ns.
		if (toBeLowered.right.balance > 0)
			toBeLowered.balance = toBeLowered.balance - 1 - toBeLowered.right.balance;
		else
			toBeLowered.balance--;
		if (toBeLowered.balance < 0)
			toBeLowered.right.balance = toBeLowered.right.balance - 1 + toBeLowered.balance;
		else
			toBeLowered.right.balance--;

		if  (toBeLowered.parent.left == toBeLowered)
			toBeLowered.parent.left = toBeLowered.right;
		else
			toBeLowered.parent.right = toBeLowered.right;

		toBeLowered.right.parent = toBeLowered.parent;

		toBeLowered.parent = toBeLowered.right;
		toBeLowered.right  = toBeLowered.right.left;
		toBeLowered.right.parent  = toBeLowered;
		toBeLowered.parent.left = toBeLowered;
	}



	protected static class AVLNode implements INodeMap
	{
		public AVLNode left = null;
		public AVLNode right = null;
		public AVLNode parent = null;
		public int balance = 0;
		public Object key;
		public Object element;


		public AVLNode()	{}
		public AVLNode(Object keyNew, Object elementNew)
		{
			key = keyNew;
			element = elementNew;
		}


		public Object getKey() {return key;}
		public void setKey(Object keyNew) {key = keyNew;}

		public Object getElement() {return element;}
		public void setElement(Object elementNew) {element = elementNew;}

		public AVLNode getLeft() {return left;}
		public void setLeft(AVLNode leftNew) {left = leftNew;}

		public AVLNode getRight() {return right;}
		public void setRight(AVLNode rightNew) {right = rightNew;}

		public AVLNode getParent() {return parent;}
		public void setParent(AVLNode parentNew) {parent = parentNew;}

		public int getBalance() {return balance;}
		public void setBalance(int balanceNew) {balance = balanceNew;}
	}

	public static class Terminator extends AVLNode	{}


	public static class TerminatorFactory
	{
		public Terminator createTerminator(AVLNode nodeForMe, boolean forLeft)
		{
			Terminator tmp = new Terminator();
			tmp.parent = nodeForMe;
			tmp.right = null;
			tmp.left = null;
			return tmp;
		}
	}
}