package rob.util;


import java.util.Comparator;


public class AVLTree
{
	// Objeto usado para comparar elementos enquanto so
	// inseridos na rvore. O elemento a ser inserido
	// sempre ser o primeiro parmetro passado para o
	// mtodo compare().
	private Comparator comp;

	// Objeto usado para decidir o que fazer quando duas
	// chaves consideradas iguais so encontradas. O mtodo
	// resolve(Node element1, Node element2)  ento
	// chamado e em caso de retorno zero a execuo da 
	// funco para de descer na rvore e parte para a
	// finalizao do mtodo. No esquecer que o Resolutor
	// deve possuir um apontador para a rvore em questo
	private Resolutor resol;
		

	// O n dessa rvore tem parent apontando para s mesmo
	// e left apontando para o primeiro elemento da rvore.
	private AVLNode head;


	private long size = 0;


	public AVLTree(Comparator compNew, Resolutor resolNew)
	{
		head = new AVLNode();
		comp = compNew; 
		resol = resolNew;
	}


	public static final byte PRE_ORDER = 1;
	public static final byte IN_ORDER = 2;
	public static final byte POS_ORDER = 3;
	public void traverse(byte mode)
	{
		if (head.left == null)
			return;
		if (mode == PRE_ORDER)
			traversePre(head.left);
		else if (mode == POS_ORDER)
			traversePos(head.left);
		else // if (mode == IN_ORDER)
			traverseIn(head.left);
	}


	protected void traversePre(AVLNode theOne)
	{
		if (theOne.left != null)
			traversePre(theOne.left);
		if (theOne.right != null)
			traversePre(theOne.right);
		System.out.print("|" + theOne.key.toString() + ", " + theOne.balance + "|");
	}


	protected void traversePos(AVLNode theOne)
	{
		System.out.print("|" + theOne.key.toString() + ", " + theOne.balance + "|");
		if (theOne.left != null)
			traversePos(theOne.left);
		if (theOne.right != null)
			traversePos(theOne.right);
	}


	protected void traverseIn(AVLNode theOne)
	{
		if (theOne.left != null)
			traverseIn(theOne.left);
		System.out.print("|" + theOne.key.toString() + ", " + theOne.balance + "|");
		if (theOne.right != null)
			traverseIn(theOne.right);
	}


	public void put(Object key, Object element)
	{
		if (head.left == null)
		{
			size++;
			head.left = new AVLNode(key, element);
			head.left.parent = head;
			return;
		}

		AVLNode nodeIterator = head.left;
		AVLNode nodeNew = new AVLNode(key, element);
		int result;
		while (true)
		{
			result = comp.compare(nodeNew.key, nodeIterator.key);

			// Decide o que fazer com o n em caso de valores de chaves iguais.
			if (result == 0)
				result = resol.resolve(nodeNew, nodeIterator);

			// Adiciona na esquerda.
			if  (result  < 0)
			{
				if  (nodeIterator.left == null)
				{
					nodeNew.parent = nodeIterator;
					nodeIterator.left = nodeNew;
					fixupInsertion(nodeNew);
					return;
				}
				else
				{
					nodeIterator = nodeIterator.left;
					continue;
				}
			}

			// Adiciona na direita.
			else if (result  > 0)
			{
System.out.print("|adio a dereta   |");
				if  (nodeIterator.right == null)
				{
					nodeNew.parent = nodeIterator;
					nodeIterator.right = nodeNew;
System.out.print("|entra fixup|");
					fixupInsertion(nodeNew);
					return;
				}
				else
				{
					nodeIterator = nodeIterator.right;
					continue;
				}
			}
			// Adiciona no mesmo.
			else
			{
				nodeNew.parent = nodeIterator.parent;
				nodeNew.left = nodeIterator.left;
				nodeNew.right = nodeIterator.right;
				if (nodeIterator.parent.left == nodeIterator)
					nodeIterator.parent.left = nodeNew;
				else
					nodeIterator.parent.right = nodeNew;
				if (nodeIterator.right != null)
					nodeIterator.right.parent = nodeNew;
				if (nodeIterator.left != null)
					nodeIterator.left.parent = nodeNew;
				nodeNew.balance = nodeIterator.balance;
				return;
			}
		}
	}


	public Object get(Object key)
	{
		AVLNode nodeIterator = head.left;
		int result = 0;
		while (nodeIterator != null)
		{
			result = comp.compare(key, nodeIterator.key);

			// Decide o que fazer com o n em caso de valores de chaves iguais.
			if (result == 0)
				result = resol.resolveKeys(key, nodeIterator.key);
 			// Desce para esquerda.
			if (result > 0)
				nodeIterator = nodeIterator.right;
			// Desce para direita.
			else if  (result < 0)
				nodeIterator = nodeIterator.left;
			// Achou!
			else
				return nodeIterator.element;
		}
		return null;
	}


	public Object remove(Object key)
	{
		AVLNode nodeIterator = head.left;
		int result = 0;
System.out.print("|entered remotion|");
		while (nodeIterator != null)
		{
			result = comp.compare(key, nodeIterator.key);
			if (result == 0)
				result = resol.resolveKeys(key, nodeIterator.key);
			if  (result > 0)
				nodeIterator = nodeIterator.right;
			else if  (result < 0)
				nodeIterator = nodeIterator.left;
			else
			{
System.out.print("|xxx/\nodeIterator.balance = " + nodeIterator.balance + "| ");
System.out.print("|xxx/\nodeIterator.parent.balance = " + nodeIterator.parent.balance + "| ");
if (nodeIterator.parent.balance == 1)
{
System.out.print("|xxx/\nodeIterator.parent.right = " + nodeIterator.parent.right + "| ");
if (nodeIterator.parent.right != null)
System.out.print("|xxx/\nodeIterator.parent.right.balance = " + nodeIterator.parent.right.balance + "| ");
}
	
				if (nodeIterator.left == null)
				{
					// n no tem filhos.
					if (nodeIterator.right == null)
					{
						if (nodeIterator.parent.left == nodeIterator)
						{
							nodeIterator.parent.left = null;
							fixupRemotion(nodeIterator.parent, 1);
						}
						else
						{
							nodeIterator.parent.right = null;
							fixupRemotion(nodeIterator.parent, -1);
						}
					}
					// s tem filho direito.
					else
					{
System.out.print("|// s tem filho direito| ");
						nodeIterator.right.parent = nodeIterator.parent;
						if (nodeIterator.parent.left == nodeIterator)
						{
System.out.print("|pai esquerdo-> nodeIterator.parent.balance = " + nodeIterator.parent.balance + "| ");
System.out.print("|nodeIterator.balance = " + nodeIterator.balance + "| ");
System.out.print("|nodeIterator.right.balance = " + nodeIterator.right.balance + "| ");
							nodeIterator.parent.left = nodeIterator.right;
							fixupRemotion(nodeIterator.parent, 1);
						}
						else
						{
System.out.print("|pai direito-> nodeIterator.parent.balance = " + nodeIterator.parent.balance + "| ");
System.out.print("|nodeIterator.balance = " + nodeIterator.balance + "| ");
System.out.print("|nodeIterator.right.balance = " + nodeIterator.right.balance + "| ");
							nodeIterator.parent.right = nodeIterator.right;
							fixupRemotion(nodeIterator.parent, -1);
						}
					}
				}
				else
				{
					// s tem filho esquerdo.
					if (nodeIterator.right == null)
					{
System.out.print("|// s tem filho esquerdo.| ");
						nodeIterator.left.parent = nodeIterator.parent;
						if (nodeIterator.parent.left == nodeIterator)
						{
							nodeIterator.parent.left = nodeIterator.left;
							fixupRemotion(nodeIterator.parent, 1);
						}
 						else
						{
							nodeIterator.parent.right = nodeIterator.left;
							fixupRemotion(nodeIterator.parent, -1);
						}
					}
					// tem os dois filhos.
					else
					{
System.out.print("|// tem os dois filhos.| ");
						if (nodeIterator.right.left == null)
						{
							if (nodeIterator.parent.left == nodeIterator)
								nodeIterator.parent.left = nodeIterator.right;
							else
								nodeIterator.parent.right = nodeIterator.right;
							nodeIterator.right.left = nodeIterator.left;
							nodeIterator.left.parent = nodeIterator.right;
							nodeIterator.right.parent = nodeIterator.parent;
							nodeIterator.right.balance = nodeIterator.balance;
							fixupRemotion(nodeIterator.right, -1);
						}
						else
						{
							AVLNode changeNode = nodeIterator.right;
							while (changeNode.left != null)
								changeNode = changeNode.left;

							changeNode.parent.left = changeNode.right;
							if (changeNode.right != null)
								changeNode.right.parent = changeNode.parent;

							AVLNode nodeTmp = changeNode.parent;

							changeNode.parent = nodeIterator.parent;
							changeNode.right = nodeIterator.right;
							changeNode.left = nodeIterator.left;
							changeNode.balance = nodeIterator.balance;

							if (nodeIterator.parent.left == nodeIterator)
								nodeIterator.parent.left = changeNode;
							else
								nodeIterator.parent.right = changeNode;
							nodeIterator.right.parent = changeNode;
							nodeIterator.left.parent = changeNode;
							fixupRemotion(nodeTmp, 1);
						}
					}
				}
System.out.print("|return nodeIterator.element|");
				return nodeIterator.element;
			}
		}
System.out.print("|return null|");
		return null;
	}


/**
 * 	D um jeito na remoo. diff diz a diferena de altura atual.
 */
	public void fixupRemotion(AVLNode toBeChanged, int diff)
	{
		size--;
		if (toBeChanged == head)
			return;
		// Para o primeiro n, a rvore vai seguir a mudana de altura.
		toBeChanged.balance += diff;
System.out.print("|entered fixupRemotion|");
		if (toBeChanged.balance == 2)
		{
System.out.print("|toBeChanged.balance == 2|");
			if  (toBeChanged.right.balance >= 0)
				shrinkRightSubTree(toBeChanged);
			else
			{
				shrinkLeftSubTree(toBeChanged.right);
				shrinkRightSubTree(toBeChanged);
			}
//////			toBeChanged.balance = 0;
				toBeChanged = toBeChanged.parent;
		}
		else if (toBeChanged.balance == -2)
		{
System.out.print("|toBeChanged.balance == -2|");
			if  (toBeChanged.left.balance <= 0)
				shrinkLeftSubTree(toBeChanged);
			else
			{
				shrinkRightSubTree(toBeChanged.left);
				shrinkLeftSubTree(toBeChanged);
			}
//////			toBeChanged.balance = 0;
				toBeChanged = toBeChanged.parent;
		}
		else
		{
System.out.print("|else toBeChanged.balance == 2/-2|");
			if (toBeChanged.balance != 0)
				return;
		}

		AVLNode nodeParent = toBeChanged.parent;
System.out.print("|entered while (nodeParent != head)|");
		while (nodeParent != head)
	  	{
			if (nodeParent.left == toBeChanged)
			{
				if (nodeParent.balance == 0)
				{
					nodeParent.balance = 1;
					return;
				}
				else if (nodeParent.balance == -1)
					nodeParent.balance = 0;
				// Rotao.
				else
				{
					nodeParent.balance = 2;
System.out.print("|going to if  (nodeParent.right.balance == 1)|");
					if  (nodeParent.right.balance >= 0)
						shrinkRightSubTree(nodeParent);
					else
					{
						shrinkLeftSubTree(nodeParent.right);
						shrinkRightSubTree(nodeParent);
					}
//////					nodeParent.balance = 0;
					nodeParent = nodeParent.parent;
				}
			}
			else
			{
				if (nodeParent.balance == -1)
				{
					nodeParent.balance = -2;
System.out.print("|going to if  (nodeParent.left.balance == -1)|");
					if (nodeParent.left.balance <= 0)
						shrinkLeftSubTree(nodeParent);
					else
					{
						shrinkRightSubTree(nodeParent.left);
						shrinkLeftSubTree(nodeParent);
					}
//////					nodeParent.balance = 0;
					nodeParent = nodeParent.parent;
				}
				else if (nodeParent.balance == 1)
					nodeParent.balance = 0;
				else
				{
					nodeParent.balance = -1;
					return;
				}
			}
			toBeChanged = nodeParent;
			nodeParent = nodeParent.parent;
		}
	}


/**
 * 	D um jeito na insero.
 */
	public void fixupInsertion(AVLNode nodeNew)
	{
		size++;
		AVLNode nodeParent = nodeNew.parent;
System.out.print ("|entered fixupInsertion|");
		while (nodeParent != head)
	  	{
			if (nodeParent.left == nodeNew)
			{
System.out.print ("|nodeParent.left == nodeNew|");
				// Se a rvore da direita j era maior do que a da esquerda, no muda de tamanho, s diminue o balano.
				if (nodeParent.balance > 0)
				{
					nodeParent.balance = 0;
					return;
				}
				// Aqui a rvore da direita  igual a da esquerda, ento muda de tamanho e desbalanceia.
				else if (nodeParent.balance == 0)
				{
System.out.print ("|nodeParent.balance == 0|");
					nodeParent.balance = -1;
				}
				// Rotao.
				else
				{
					nodeParent.balance = -2;
System.out.print ("|nodeParent.left.balance = " +nodeParent.left.balance + "|");
					if  (nodeParent.left.balance == -1)
						shrinkLeftSubTree(nodeParent);
					else
					{
						shrinkRightSubTree(nodeParent.left);
						shrinkLeftSubTree(nodeParent);
					}
///////					nodeParent.balance = 0;
					nodeParent = nodeParent.parent;
					return ;
				}
			}
			else
			{
				// Se a rvore da direita j era maior do que a da esquerda, ento aumenta a rvore e a rotaciona.
				if (nodeParent.balance > 0)
				{
					nodeParent.balance = 2;
System.out.print ("|nodeParent.balance > 0|");
					if  (nodeParent.right.balance == 1)
					{
System.out.print("|shrinkRightSubTree(nodeParent)|");
						shrinkRightSubTree(nodeParent);
					}
					else
					{
System.out.print("|shrinkLeftSubTree(nodeParent.right)->shrinkRightSubTree(nodeParent)"+nodeParent.right.balance+"|");
						shrinkLeftSubTree(nodeParent.right);
						shrinkRightSubTree(nodeParent);
					}
//////					nodeParent.balance = 0;
					nodeParent = nodeParent.parent;
					return ;
				}
				// Aqui a rvore da direita  igual a da esquerda, ento
				//  muda de tamanho e desbalanceia.
				else if (nodeParent.balance == 0)
				{
System.out.print("|nodeParent.balance == 0|");
					nodeParent.balance = 1;
				}
				// Zera o balance e para o algoritmo.
				else
				{
					nodeParent.balance = 0;
					return ;
				}

			}
			// Caminha em direo do prximo do topo da rvore.
			nodeNew = nodeParent;
			nodeParent = nodeParent.parent;
		}
	}


	public void shrinkLeftSubTree(AVLNode toBeLowered)
	{
		// Ajusta o balano dos ns.
		if (toBeLowered.left.balance < 0)
			toBeLowered.balance = toBeLowered.balance + 1 - toBeLowered.left.balance;
		else
			toBeLowered.balance++;;
		if (toBeLowered.balance > 0)
			toBeLowered.left.balance = toBeLowered.left.balance + 1 + toBeLowered.balance;
		else
			toBeLowered.left.balance++;

System.out.print ("|entered shrinkLeftSubTree|");

		// Cuida de arrumar o pai de todos (no t na jogada).
		if  (toBeLowered.parent.left == toBeLowered)
			toBeLowered.parent.left = toBeLowered.left;
		else
			toBeLowered.parent.right = toBeLowered.left;

		toBeLowered.left.parent = toBeLowered.parent;

		toBeLowered.parent = toBeLowered.left;
		toBeLowered.left  = toBeLowered.left.right;
		if (toBeLowered.left != null)
			toBeLowered.left.parent = toBeLowered;
		toBeLowered.parent.right = toBeLowered;
	}


	public void shrinkRightSubTree(AVLNode toBeLowered)
	{
		// Ajusta o balano dos ns.
		if (toBeLowered.right.balance > 0)
			toBeLowered.balance = toBeLowered.balance - 1 - toBeLowered.right.balance;
		else
			toBeLowered.balance--;
		if (toBeLowered.balance < 0)
			toBeLowered.right.balance = toBeLowered.right.balance - 1 + toBeLowered.balance;
		else
			toBeLowered.right.balance--;

System.out.print ("|entered shrinkRightSubTree|");
		if  (toBeLowered.parent.left == toBeLowered)
			toBeLowered.parent.left = toBeLowered.right;
		else
			toBeLowered.parent.right = toBeLowered.right;

		toBeLowered.right.parent = toBeLowered.parent;

		toBeLowered.parent = toBeLowered.right;
		toBeLowered.right  = toBeLowered.right.left;
		if (toBeLowered.right != null)
			toBeLowered.right.parent  = toBeLowered;
		toBeLowered.parent.left = toBeLowered;
	}



	protected static class AVLNode implements INodeMap
	{
		public AVLNode left = null;
		public AVLNode right = null;
		public AVLNode parent = null;
		public int balance = 0;
		public Object key;
		public Object element;


		public AVLNode()	{}
		public AVLNode(Object keyNew, Object elementNew)
		{
			key = keyNew;
			element = elementNew;
		}


		public Object getKey() {return key;}
		public void setKey(Object keyNew) {key = keyNew;}

		public Object getElement() {return element;}
		public void setElement(Object elementNew) {element = elementNew;}

		public AVLNode getLeft() {return left;}
		public void setLeft(AVLNode leftNew) {left = leftNew;}

		public AVLNode getRight() {return right;}
		public void setRight(AVLNode rightNew) {right = rightNew;}

		public AVLNode getParent() {return parent;}
		public void setParent(AVLNode parentNew) {parent = parentNew;}

		public int getBalance() {return balance;}
		public void setBalance(int balanceNew) {balance = balanceNew;}
	}
}