package rob.util;

import java.util.*;


/**
 *	Essa clase implementa uma lista duplamente ligada circular. Essa classe aceita todas as operaes de List. No caso de se 
 * criar um Iterator ou ListIterator, as operaes de add, addAll e remove, removeAll s podem ser realizadas se apenas um dos
 * iteratos for matido, os demias passaro a jogar exceo ConcurrentSomethingException.
 * 	A lista apresenta ainda algumas outras funes interessantes, alm de adaptar um pouco as definies de algumas funes 
 * da interface List, pois essa lista no tem fim.
 */

///////public class CDLinkedList extends AbstractSequentialList implements List, ListIterator, Cloneable, java.io.Serializable
///////public abstract class AbstractCDLList implements List, Cloneable, java.io.Serializable
public class AbstractCDLList implements List, Cloneable, java.io.Serializable
{
 	protected transient Entry head;
	protected int size;
	protected transient long modificationCounter;


/**
 * 	Constructs an empty list.
 */
//////	protected AbstractCDLList()
	public AbstractCDLList()
	{
		head = null;
		size = 0;
		modificationCounter = 0;
	}


/**
 * 	Constructs a list containing the elements of the specified
 * collection, in the order they are returned by the collection's
 * iterator.
 */
//////	protected AbstractCDLList(Collection c) 
	public AbstractCDLList(Collection c) 
	{
		this();
		System.out.print("  Entrou... ");
		addAll(c);
	}

/*
 *	Adiciona um elemento antes de Entry. Se for passado o head, adiciona na
 * posio 0, e se for passado head.previous adiciona na posio size - 2, e se
 * for passado head.next, adiciona na posio 1. (-2, 0, 1)
 */
	protected Entry addBefore(Entry h, Object toBeAdded)
	{
		modificationCounter++;
		size++;
		if (h == null)	// Lista ainda vaiza ...
		{
			h = new Entry(toBeAdded, h, h);
			h.previous = h; h.next = h;
			head = h;
			return h;
		}
		else
		{
			Entry tmpEntry = new Entry(toBeAdded, h, h.previous);
			h.previous.next = tmpEntry;
			h.previous = tmpEntry;
			if (head == h) 	// Se for o primeiro elemento ...
				head = h.previous;
			return tmpEntry;
		}
	}


/*
 *	Adiciona um elemento depois de Entry. Se for passado o head, adiciona na
 * posio 1, e se for passado head.previous adiciona na posio size -1, e se
 * for passado head.next adiciona na posio 2. (-1, 1, 2)
 */
	protected Entry addAfter(Entry h, Object toBeAdded)
	{
		modificationCounter++;
		size++;
		if (h == null)	// Lista ainda vaiza ...
		{
			h = new Entry(toBeAdded, h, h);
			h.previous = h; h.next = h;
			head = h;
			return h;
		}
		else
		{
			Entry tmpEntry = new Entry(toBeAdded, h.next, h);
			h.next.previous = tmpEntry;
			h.next = tmpEntry;
			return tmpEntry;
		}
	}


/*
 *	Remove o elemento depois Entry.
 */
	protected Entry remove(Entry h)
	{
		if (h == null)	// Lista ainda vaiza ...
			throw new NullPointerException("Lista vazia bixo!");
		else
		{
			modificationCounter++;
			size--;
			if (h == h.next)
			{
				head = null;
				return null;
			}
			h.next.previous = h.previous;
			h.previous.next = h.next;
			if (head == h)
				head = h.next;
			return h;
		}
	}


/**
 *	Move head X ndices.
 */
	protected int moveHead(int index)
	{
		if (head == null)
			return 0;
		index = (index >= 0) ? index%size : size - (-index)%size;
		int howMuch = index;
		if (index > size - index)
		{
			index = size - index;
			while (--index > -1)
				head = head.previous;
		}
		else
		{
			while (--index > -1)
				head = head.next;
		}
		return howMuch;
	}


/**
 *	Move entry X ndices.
 */
	protected Entry moveEntry(int index, Entry entry)
	{
		if (entry == null)
			return null;
		index = (index >= 0) ? index%size : size - (-index)%size;
		int howMuch = index;
		if (index > size - index)
		{
			index = size - index;
			while (--index > -1)
				entry = entry.previous;
		}
		else
		{
			while (--index > -1)
				entry = entry.next;
		}
		return entry;
	}


	protected Entry moveEntry(Entry entry, int index)
	{
		if (entry == null)
			return null;
		index = (index >= 0) ? index%size : size - (-index)%size;
		int howMuch = index;
		if (index > size - index)
		{
			index = size - index;
			while (--index > -1)
				entry = entry.previous;
		}
		else
		{
			while (--index > -1)
				entry = entry.next;
		}
		return entry;
	}


	protected int adjustToPositiveInBoundsValue(int indexOld)
	{
		return (indexOld + size)%size;
	}


/////////////////////	Interface List.


/**
 * Appends the specified element to the pevious element of the header.
 * @param o element to be appended to this list.
 * @return <tt>true</tt> (as per the general contract of
 * <tt>Collection.add</tt>).
 */
	public boolean add(Object o)
	{
		addBefore(head, o);
		return true;
	}


/**
 * Appends the specified element to the specified index.
 * @param o element to be appended to this list.
 * @return <tt>true</tt> (as per the general contract of
 * <tt>Collection.add</tt>).
 */
	public void add(int index, Object o) 
	{
		if (index == 0)
			addBefore(head, o);
		else
		{
			Entry saveHead = head;
			moveHead(index - 1);
			addAfter(head, o);
			head = saveHead;
		}
	}


/**
 * Appends all the specified element to the previous element of the header.
 * @param o element to be appended to this list.
 * @return <tt>true</tt> (as per the general contract of
 * <tt>Collection.add</tt>).
 */
	public boolean addAll(Collection c) 
	{
		boolean flag = false;
		if (head == null)
		{
			flag = true;
			addBefore(head, new String("X"));
		}
		Iterator itColl = c.iterator();
		while (itColl.hasNext())
			addAfter(head.previous, itColl.next());
		if (flag)
			remove(head);
		return true;
	}


/**
 * Appends all the specified element to the next element of the header
 * starting index elements of the head.
 *	This method throws UnsupportedOperationException() if some
 * Iterator is active on this List.
 * @param o element to be appended to this list.
 * @return <tt>true</tt> (as per the general contract of
 * <tt>Collection.add</tt>).
 */
	public boolean addAll(int index, Collection c) 
	{

		boolean flag = false;
		if (head == null)
		{
			flag = true;
			addBefore(head, new String("X"));
		}
		Entry saveHead = head;
		moveHead(index);
		addAll(c);
		head = saveHead;
		if (flag)
			remove(head);
		return true;
	}


/**
 * Removes all of the elements from this list.
 */
 	public void clear() 
	{
		head = null;
		size = 0;
		modificationCounter++;
	}


/**
 * Returns <tt>true</tt> if this list contains the specified element.
 * More formally, returns <tt>true</tt> if and only if this list contains
 * at least one element <tt>e</tt> such that <tt>(o==null ? e==null
 * : o.equals(e))</tt>.
 * 
 * @param o element whose presence in this list is to be tested.
 * @return <tt>true</tt> if this list contains the specified element.
 */
 	public boolean contains(Object o) 
	{
		return indexOf(o) != -1;
	}


/**
 * Returns <tt>true</tt> if this list contains the set of specified elements.
 * More formally, returns <tt>true</tt> if and only if this list contains
 * at least one of each collectoin's elements <tt>e</tt> such that <tt>(o==null ? 
 * e==null : o.equals(e))</tt>.
 * 
 * @param o element whose presence in this list is to be tested.
 * @return <tt>true</tt> if this list contains the specified element.
 */
 	public boolean containsAll(Collection c) 
	{
		Iterator itColl = c.iterator();
		while(itColl.hasNext())
			if (!contains(itColl.next()))
				return false;
		return true;
	}


/**
 * Compares the specified object with this list for equality.  Returns
 * <tt>true</tt> if and only if the specified object is also a list, both
 * lists have the same size, and all corresponding pairs of elements in
 * the two lists are <i>equal</i>.  (Two elements <tt>e1</tt> and
 * <tt>e2</tt> are <i>equal</i> if <tt>(e1==null ? e2==null :
 * e1.equals(e2))</tt>.)  In other words, two lists are defined to be
 * equal if they contain the same elements in the same order.<p>
 *
 * This implementation first checks if the specified object is this
 * list. If so, it returns <tt>true</tt>; if not, it checks if the
 * specified object is a list. If not, it returns <tt>false</tt>; if so,
 * it iterates over both lists, comparing corresponding pairs of elements.
 * If any comparison returns <tt>false</tt>, this method returns
 * <tt>false</tt>.  If either iterator runs out of elements before the
 * other it returns <tt>false</tt> (as the lists are of unequal length);
 * otherwise it returns <tt>true</tt> when the iterations complete.
 *
 * @param o the object to be compared for equality with this list.
 * 
 * @return <tt>true</tt> if the specified object is equal to this list.
 */
	public boolean equals(Object o) 
	{
		if (o == null || !(o instanceof List))
			return false;
		if (o == this)
			return true;
		ListIterator e1 = listIterator();
		ListIterator e2 = ((List) o).listIterator();
		Object o1 = null;
		Object o2 = null;
		while(e1.hasNext() && e2.hasNext())
		{
			o1 = e1.next(); o2 = e2.next();
			if (o1 == null || !o1.equals(o2))
				return false;
		}
		return !(e1.hasNext() || e2.hasNext());
	}


/**
 * Returns the element at index elements of the head.
 *
 * @param index index of element to return.
 * @return the element at the specified position in this list.
 * 
 * @throws IndexOutOfBoundsException if the specified index is is out of
 * range (<tt>index &lt; 0 || index &gt;= size()</tt>).
 */
	public Object get(int index)
	{
		if (head == null)
			throw new NoSuchElementException("Empty list.");
		Entry saveHead = head;
		moveHead(index);
		Object retValue = head.element;
		head = saveHead;
		return retValue;
		}


/**
 * Returns the hash code value for this list. <p>
 *
 * This implementation uses exactly the code that is used to define the
 * list hash function in the documentation for the <tt>List.hashCode</tt>
 * method.
 *
 * @return the hash code value for this list.
 */
	public int hashCode() 
	{
		int hashCode = 1;
		Iterator i = iterator();
		while (i.hasNext()) 
		{
			Object obj = i.next();
			hashCode = 31*hashCode + (obj == null ? 0 : obj.hashCode());
		}
		return hashCode;
	}


/**
 * Returns the index in this list of the first occurrence of the
 * specified element, or -1 if the List does not contain this
 * element.  More formally, returns the lowest index i such that
 * <tt>(o==null ? get(i)==null : o.equals(get(i)))</tt>, or -1 if
 * there is no such index.
 *
 * @param o element to search for.
 * @return the index in this list of the first occurrence of the
 * 	       specified element, or -1 if the list does not contain this
 * 	       element.
 */
	public int indexOf(Object o) 
	{
		if (head == null)
			throw new NoSuchElementException("Empty list.");
		int index = -1;
		Entry headTmp = head;
		while (++index < size)
		{
			if (headTmp.element.equals(o))
				return index;
			headTmp = headTmp.next;
		}
		return -1;
	}


	public boolean isEmpty()
	{
		return size == 0 ? true : false;
	}


	public Iterator iterator()
	{		
		return listIterator();
	}


/**
 * Returns the index in this list of the last occurrence of the
 * specified element, or -1 if the list does not contain this
 * element.  More formally, returns the highest index i such that
 * <tt>(o==null ? get(i)==null : o.equals(get(i)))</tt>, or -1 if
 * there is no such index.
 *
 * @param o element to search for.
 * @return the index in this list of the last occurrence of the
 * 	       specified element, or -1 if the list does not contain this
 * 	       element.
 */
	public int lastIndexOf(Object o) 
	{
		if (head == null)
			throw new NoSuchElementException("No next element.");
		int index = -1;
		int theIndex = -1;
		Entry headTmp = head;
		while (++index < size)
		{
			if (headTmp.element.equals(o))
				theIndex = index;
			headTmp = headTmp.next;
		}
		return theIndex;
	}


/**
 *	Mtodo que retorna um ListIterator para esta Lista. Quando um Iterator
 *  criado, as operae de add e remove ficam bloqueadas. Ento toda vez que
 * um Iterator  usado (e no ir se usa-lo mais), deve-se chamar disposeIterator
 * ou na List de onde ele veio ou sobre ele mesmo.
 */
	public ListIterator listIterator()
	{	
		return new CDLListIterator();;
	}


	public ListIterator listIterator(int index)
	{
		Entry saveHead = head;
		moveHead(index);
		ListIterator retVal = new CDLListIterator();
		head = saveHead;
		return retVal;
	}


	public Object remove(int index)
	{
		Entry saveHead = head;
		moveHead(index);
		remove(head);
		Object retVal = head.element;
		head = saveHead;
		return retVal;
	}


/**
 * Removes the first occurrence of the specified element in this list.  If
 * the list does not contain the element, it is unchanged.  More formally,
 * removes the element with the lowest index <tt>i</tt> such that
 * <tt>(o==null ? get(i)==null : o.equals(get(i)))</tt> (if such an
 * element exists).
 *
 * @param o element to be removed from this list, if present.
 * @return <tt>true</tt> if the list contained the specified element.
 */
	public boolean remove(Object o) 
	{
		if (head == null)
			return false;
		Entry headIt = head;
		int i = -1;
		while (++i < size)
		{
			if (headIt.element.equals(o))
			{
				remove(headIt);
				return true;
			}
			headIt = headIt.next;
		}
		return false;
	}


/**
 * Removes the first occurrence of the specifieds elements in this list.  If
 * the list does not contain none of the elements, it is unchanged.  More formally,
 * removes the element with the lowest index <tt>i</tt> such that
 * <tt>(o==null ? get(i)==null : o.equals(get(i)))</tt> (if such an
 * element exists).
 *
 * @param o element to be removed from this list, if present.
 * @return <tt>true</tt> if the list contained the specified element.
 */
	public boolean removeAll(Collection c) 
	{
		Iterator itColl = c.iterator();
		boolean removed = false;
		while (itColl.hasNext())
			if (remove(itColl.next()))
				removed = true;
		return removed;
	}


	public boolean retainAll(Collection c)
	{
		throw new UnsupportedOperationException("retainAll");
	}


/**
 * Replaces the element at the specified position in this list with the
 * specified element.
 *
 * @param index index of element to replace.
 * @param element element to be stored at the specified position.
 * @return the element previously at the specified position.
 * @throws IndexOutOfBoundsException if the specified index is out of
 *		  range (<tt>index &lt; 0 || index &gt;= size()</tt>).
 */
	public Object set(int index, Object element)
	{
		if (head == null)
		{
			add(element);
			return null;
		}
		Entry saveHead = head;
		moveHead(index);
		Object retVal;
		retVal = head.element;
		head.element = element;
		head = saveHead;
		return retVal;
	}


/**
 * Returns the number of elements in this list.
 *
 * @return the number of elements in this list.
 */
	public int size() 	{return size;}


/**
 *	Retorna uma lista indo de fromIndex at toIndex, sendo contados a partir
 * de head, incluindo fromIndex e excluindo toIndex.
 */
	public List subList(int fromIndex, int toIndex)
	{
		List retVal = new AbstractCDLList();
		Entry saveHead = head;
		ListIterator itList = listIterator();
		fromIndex = moveHead(fromIndex);
		toIndex = adjustToPositiveInBoundsValue(toIndex);
		toIndex = (toIndex < fromIndex) ? fromIndex - toIndex : toIndex - fromIndex;
		int i = -1;
		while (++i < toIndex)
			retVal.add(itList.next());
		return retVal;
	}


	public Object[] toArray() 
	{
		Object[] result = new Object[size];
		int i = -1;
		ListIterator itList = listIterator();
		while (++i < size)
			result[i] = itList.next();
		return result;
	}


	public Object[] toArray(Object a[])
	{
		if (a.length < size)
		a = (Object[]) java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size);
		int i = -1;
		ListIterator itList = listIterator();
		while (++i < size)
			a[i] = itList.next();
		if (a.length > size)
			a[size] = null;
		return a;
	}


/////////////////////	Interface ListIterator.


/**
 *	A classe CDListIt mantm um sistema de ndice baseado no no nmero de elementos
 * que a lista a que  associada tem comeando na posio que o head da List estava na
 * criano da instncia atual dessa classe.
 */
	protected class CDLListIterator implements ListIterator
	{
	 	protected transient Entry head2;
		protected transient int itIndex;
		protected Entry lastReturned;
		protected long localModificationCounter;


		public CDLListIterator()
		{
			if (head == null)
			{
				head2 = new Entry(null, null, null);
				head2.next = head2; head2.previous = head2;
				head2.element = null;
			}
			else
				head2 = head.previous;
			lastReturned = null;
			itIndex = -1;
			localModificationCounter = AbstractCDLList.this.modificationCounter;
		}


/**
 *	Still not supportted.
 */
		public void add(Object o) 
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			AbstractCDLList.this.addAfter(lastReturned.previous, o);
			itIndex++;
			localModificationCounter++;
		}


		public boolean hasNext()
		{
			return (itIndex == size - 1) ? false : true;
		}


		public boolean hasPrevious()
		{
			return (itIndex == -1) ? false : true;
		}


		public Object next()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			if (!hasNext())
				throw new NoSuchElementException("No next element.");
			head2 = head2.next;
			itIndex++;
			lastReturned = head2;
			return lastReturned.element;
		}


		public int nextIndex()
		{
			return itIndex + 1;
		}


		public Object previous()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			if (!hasPrevious())
				throw new NoSuchElementException("No previous element.");
			itIndex--;
			lastReturned = head2;
			head2 = head2.previous;
			return lastReturned.element;
		}


		public int previousIndex()
		{
			return itIndex;
		}


/**
 *	Still not supportted.
 */
		public void remove()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			AbstractCDLList.this.remove(lastReturned);
			itIndex--;
			localModificationCounter++;
		}


/**
 * Troca o o ltimo elemento devolvido por next ou previous pelo
 * elemento element. Faz parte da interface ListIterator.
 * specified element.
 *
 * @param index index of element to replace.
 * @param element element to be stored at the specified position.
 * @return the element previously at the specified position.
 * @throws IndexOutOfBoundsException if the specified index is out of
 *		  range (<tt>index &lt; 0 || index &gt;= size()</tt>).
 */
		public void set(Object element)
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			if (lastReturned != null)
				lastReturned.element = element;
			else
				throw new NoSuchElementException("No next/previous element returned.");
		}
		
		protected boolean wasModified()
		{
			return localModificationCounter != AbstractCDLList.this.modificationCounter;
		}
	}

/////////////////////	Interface object.


/**
 *	Clona esse objeto.
 */
	public Object clone()
	{
		return new AbstractCDLList(this);
	}


/////////////////////	Interface Serializable.


	private static final long serialVersionUID = 876323262645176355L; // 876323262645176354L;

/**
 * Save the state of this <tt>LinkedList</tt> instance to a stream (that
 * is, serialize it).
 *
 * @serialData The size of the list (the number of elements it
 *		   contains) is emitted (int), followed by all of its
 * elements (each an Object) in the proper order.  
 */
	private synchronized void writeObject(java.io.ObjectOutputStream s) throws java.io.IOException
	{
		s.defaultWriteObject();
		s.writeInt(size);
		for (Entry e = head.next; e != head; e = e.next)
	            s.writeObject(e.element);
	}

/**
 * Reconstitute this <tt>LinkedList</tt> instance from a stream (that is
 * deserialize it).
 */
 	private synchronized void readObject(java.io.ObjectInputStream s) throws java.io.IOException, ClassNotFoundException 
	{
		s.defaultReadObject();
		int size = s.readInt();
		head = new Entry(null, null, null);
		head.next = head.previous = head;
		for (int i = 0; i < size; i++)
            	add(s.readObject());
	}


/////////////////////	mtodos prprios da classe.


/**
 *	Retorna uma lista indo de fromIndex at toIndex, sendo contados a partir
 * de head, incluindo fromIndex e toIndex, e uma outra contendo os 
 * elementos restantes mais fromIndex e toIndex.
 */
	public void splitPlus(int fromIndex, int toIndex, AbstractCDLList oneSide, AbstractCDLList otherSide)
	{
		oneSide.clear();
		otherSide.clear();
		ListIterator itList = listIterator();
		int i = - 1;
		while (++i < fromIndex)
			otherSide.add(itList.next());
		itList.previous();
		i--;
		while (++i < toIndex)
			oneSide.add(itList.next());
		itList.previous();
		i--;
		while (++i < size)
			otherSide.add(itList.next());
	}

/*
 *	Classe representa um n da lista.
 */
	protected class Entry
	{
		public Object element;
		public Entry next;
		public Entry previous;

		Entry(Object element, Entry next, Entry previous) 
		{
		    this.element = element;
		    this.next = next;
		    this.previous = previous;
		}
	}

}
