package rob.util;


import javax.swing.JPanel;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.MemoryImageSource;
import java.awt.image.PixelGrabber;
import java.awt.image.Raster;
import java.awt.image.RasterFormatException;
import java.awt.image.WritableRaster;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.net.URL;
import rob.exceptions.*;


public class ImageUtil 
{
	private static final Component sComponent = new Component() {};
	private static final MediaTracker sTracker = new MediaTracker(sComponent);
	private static int sID = 0;
  

	public static boolean waitForImage(Image image) 
	{
		int id;
		synchronized(sComponent) 		{id = sID++;}
		sTracker.addImage(image, id);
		try 						{sTracker.waitForID(id);}
		catch (InterruptedException ie) 	{return false;}
		if (sTracker.isErrorID(id)) 		return false;
		return true;
	}    


	public static Image blockingLoad(String path)
	{
		Image image = Toolkit.getDefaultToolkit().getImage(path);
		if (waitForImage(image) == false) return null;
		return image;	
	}
  

	public static Image loadImage(String path)
	{
		Image image = Toolkit.getDefaultToolkit().getImage(path);
		if (waitForImage(image) == false) return null;
		return image;	
	}
  

	public static Image blockingLoad(URL url)
	{
		Image image = Toolkit.getDefaultToolkit().getImage(url);
		if (waitForImage(image) == false) return null;
		return image;
	}

  
	public static BufferedImage getBufferedImage(Image image)	{return getBufferedImage(image, BufferedImage.TYPE_INT_RGB);}
  

	public static BufferedImage getBufferedImage(Image image, int imageType) 
	{
		if (waitForImage(image) == false) return null;
		BufferedImage bufferedImage = new BufferedImage(image.getWidth(null), image.getHeight(null), imageType);
		Graphics2D gr2 = bufferedImage.createGraphics();
		gr2.drawImage(image, null, null);
		return bufferedImage;
	}


	public static BufferedImage getBufferedImage(int width, int height)
	{
		BufferedImage imgTmp = (BufferedImage) new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		if (imgTmp == null)
			throw new ImageNotCreatedException("Can't create imgOffscreen.");
		return imgTmp;
	}


	public static BufferedImage getBufferedImage(int width, int height, int typeColor)
	{
		BufferedImage imgTmp = (BufferedImage) new BufferedImage(width, height, typeColor);
		if (imgTmp == null)
			throw new ImageNotCreatedException("Can't create imgOffscreen.");
		return imgTmp;
	}


	public BufferedImage getBufferedImage(Component cmpOwner)
	{
		Rectangle2D rectTmp = cmpOwner.getBounds();
		BufferedImage imgTmp = (BufferedImage) cmpOwner.createImage((int) rectTmp.getWidth(), (int) rectTmp.getHeight());
		if (imgTmp == null)
			throw new ImageNotCreatedException("Can't create imgOffscreen.");
		return imgTmp;
	}


	public BufferedImage getBufferedImage(Component cmpOwner, int width, int height)
	{
		BufferedImage imgTmp = (BufferedImage) cmpOwner.createImage(width, height);
		if (imgTmp == null)
			throw new ImageNotCreatedException("Can't create imgOffscreen.");
		return imgTmp;
	}


	public BufferedImage getBufferedImage(Component cmpOwner, MemoryImageSource imgSource)
	{
		BufferedImage imgTmp = (BufferedImage) cmpOwner.createImage(imgSource);
		if (imgTmp == null)
			throw new ImageNotCreatedException("Can't create imgOffscreen.");
		return imgTmp;
	}


	public static BufferedImage getNewBufferedImage(BufferedImage image, int width, int height)
	{
		BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		Graphics2D gr2 = bufferedImage.createGraphics();
		gr2.drawImage(image, null, null);
		return bufferedImage;
	}

	
/**
 *	Mtodo guarda em dstArray os pixels da imagem no formato sRGB pondo todas as bandas em um s int. Se
 * no for possvel realizar tal operao,  jogada ImageNotCreatedException.  retornado o nmero de bandas
 * capatadas.  
 */
	public static int getPackedArrayOfPixels(int [] arrayDst, BufferedImage imgSrc)
	{
		int width  = imgSrc.getWidth(null);	int height = imgSrc.getHeight(null);
		if (arrayDst.length < width*height)
			throw new ImageNotCreatedException("arrayDst too small.");
		PixelGrabber pg = new PixelGrabber(imgSrc, 0, 0, width, height, arrayDst, 0, width);
		try						{pg.grabPixels();}
		catch(InterruptedException exc)	{throw new ImageNotCreatedException("Problems grabbing pixels.");}
		return 4;
	}


	public static BufferedImage getImageWithPackedArrayOfPixels(int [] arraySrc, BufferedImage imgSrc)
	{
		int width  = imgSrc.getWidth(null);	int height = imgSrc.getHeight(null);
		ColorModel colorModel = ColorModel.getRGBdefault();
		DataBuffer data = new DataBufferInt(arraySrc, arraySrc.length);
		int [] bandOffsets = new int[4];
		bandOffsets[0] = 0x00ff0000;
		bandOffsets[1] = 0x0000ff00;
		bandOffsets[2] = 0x000000ff;
		bandOffsets[3] = 0xff000000;
		WritableRaster raster = Raster.createPackedRaster(data, width, height, width, bandOffsets, null);
		try
		{
			return new BufferedImage(colorModel, raster, false, null);
		}
		catch(RasterFormatException exp1) 	{System.out.print(" Maka!! RasterFormatException -> msg: " + exp1.getLocalizedMessage()); return null;}
		catch(Exception exp1) 			{System.out.print(" Maka!! Exception -> msg: " + exp1.getLocalizedMessage()); return null;}
/*
		int width  = imgSrc.getWidth(null); int height = imgSrc.getHeight(null);
		if (arraySrc.length < width*height)
			throw new ImageNotCreatedException("arraySrc too small.");
		try
		{
			MemoryImageSource imgSource = new MemoryImageSource(width, height, arraySrc, 0, width);
			return (BufferedImage) (new JPanel()).createImage(imgSource);
		}
		catch(NullPointerException exc1)	{throw new ImageNotCreatedException("OffScreen Graphics couldn't be created.");}
*/
	}


	public static BufferedImage getBufferedImageCopy(BufferedImage imgBuffSrc)
	{
		BufferedImage imgRetVal = new BufferedImage(imgBuffSrc.getWidth(), imgBuffSrc.getHeight(), imgBuffSrc.getType());
		Graphics2D gr = imgRetVal.createGraphics();
		gr.drawImage(imgBuffSrc, null, null);
		gr.dispose();
		return imgRetVal;
	}


	public static BufferedImage getBufferedImageCopy(BufferedImage imgBuffSrc, Shape shpClip)
	{
		if (shpClip == null)
			return ImageUtil.getBufferedImageCopy(imgBuffSrc);
		BufferedImage imgRetVal = new BufferedImage(imgBuffSrc.getWidth(), imgBuffSrc.getHeight(), imgBuffSrc.getType());
		Graphics2D gr = imgRetVal.createGraphics();
		gr.clip(shpClip);
		gr.drawImage(imgBuffSrc, null, null);
		Rectangle2D rectTmp = shpClip.getBounds2D();
		gr.dispose();
		return imgRetVal.getSubimage((int) rectTmp.getX(), (int) rectTmp.getY(), (int) rectTmp.getWidth(), (int) rectTmp.getHeight());
	}


	public static void copyBufferedImage(BufferedImage imgBuffSrc, BufferedImage imgBuffDst)
	{
		Graphics2D gr = imgBuffDst.createGraphics();
		gr.drawImage(imgBuffSrc, null, null);
		gr.dispose();
	}


	public static void copyBufferedImage(BufferedImage imgBuffSrc, BufferedImage imgBuffDst, Shape shpClip)
	{
		if (shpClip == null)
			ImageUtil.copyBufferedImage(imgBuffSrc, imgBuffDst);
		else
		{
			Graphics2D gr = imgBuffDst.createGraphics();
			gr.clip(shpClip);
			gr.drawImage(imgBuffSrc, null, null);
			gr.dispose();
		}
	}
}